import json
from urllib.parse import urlencode

from .oauth import OAuth2Test


class DropboxOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.dropbox.DropboxOAuth2V2'
    user_data_url = 'https://api.dropboxapi.com/2/users/get_current_account'
    user_data_url_post = True
    expected_username = 'dbidAAH4f99T0taONIb-OurWxbNQ6ywGRopQngc'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'account_id': 'dbid:AAH4f99T0taONIb-OurWxbNQ6ywGRopQngc',
        'name': {
            'given_name': 'Franz',
            'surname': 'Ferdinand',
            'familiar_name': 'Franz',
            'display_name': 'Franz Ferdinand (Personal)',
            'abbreviated_name': 'FF'
        },
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
XING OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/xing.html
"""
from requests_oauthlib import OAuth1
from oauthlib.oauth1 import SIGNATURE_TYPE_AUTH_HEADER

from .oauth import BaseOAuth1
from ..exceptions import AuthTokenError


class XingOAuth(BaseOAuth1):
    """Xing OAuth authentication backend"""
    name = 'xing'
    AUTHORIZATION_URL = 'https://api.xing.com/v1/authorize'
    REQUEST_TOKEN_URL = 'https://api.xing.com/v1/request_token'
    ACCESS_TOKEN_URL = 'https://api.xing.com/v1/access_token'
    SCOPE_SEPARATOR = '+'
    EXTRA_DATA = [
        ('id', 'id'),
        ('user_id', 'user_id')
    ]

    def get_user_details(self, response):
        """Return user details from Xing account"""
        email = response.get('email', '')
        fullname, first_name, last_name = self.get_user_names(
            first_name=response['first_name'],
            last_name=response['last_name']
        )
        return {'username': first_name + last_name,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name,
                'email': email}

    def clean_oauth_auth(self, access_token):
        """Override of oauth_auth since Xing doesn't like callback_uri
        and oauth_verifier on authenticated API calls"""
        key, secret = self.get_key_and_secret()
        resource_owner_key = access_token.get('oauth_token')
        resource_owner_secret = access_token.get('oauth_token_secret')
        if not resource_owner_key:
            raise AuthTokenError(self, 'Missing oauth_token')
        if not resource_owner_secret:
            raise AuthTokenError(self, 'Missing oauth_token_secret')
        return OAuth1(key, secret,
                      resource_owner_key=resource_owner_key,
                      resource_owner_secret=resource_owner_secret,
                      signature_type=SIGNATURE_TYPE_AUTH_HEADER)

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        profile = self.get_json(
            'https://api.xing.com/v1/users/me.json',
            auth=self.clean_oauth_auth(access_token)
        )['users'][0]
        return {
            'user_id': profile['id'],
            'id': profile['id'],
            'first_name': profile['first_name'],
            'last_name': profile['last_name'],
            'email': profile['active_email']
        }



"""
Dropbox OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/dropbox.html
"""

from .oauth import BaseOAuth2

class DropboxOAuth2V2(BaseOAuth2):
    name = 'dropbox-oauth2'
    ID_KEY = 'uid'
    AUTHORIZATION_URL = 'https://www.dropbox.com/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://api.dropboxapi.com/oauth2/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False

    def get_user_details(self, response):
        """Return user details from Dropbox account"""
        name = response.get('name')
        return {'username': str(response.get('account_id')),
                'email': response.get('email'),
                'fullname': name.get('display_name'),
                'first_name': name.get('given_name'),
                'last_name': name.get('surname')}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://api.dropboxapi.com/2/users/get_current_account',
            headers={'Authorization': 'Bearer {0}'.format(access_token)},
            method='POST'
        )



"""
LiveJournal OpenId backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/livejournal.html
"""
from urllib.parse import urlsplit

from .open_id import OpenIdAuth
from ..exceptions import AuthMissingParameter


class LiveJournalOpenId(OpenIdAuth):
    """LiveJournal OpenID authentication backend"""
    name = 'livejournal'

    def get_user_details(self, response):
        """Generate username from identity url"""
        values = super().get_user_details(response)
        values['username'] = values.get('username') or \
            urlsplit(response.identity_url).netloc.split('.', 1)[0]
        return values

    def openid_url(self):
        """Returns LiveJournal authentication URL"""
        if not self.data.get('openid_lj_user'):
            raise AuthMissingParameter(self, 'openid_lj_user')
        return 'http://{0}.livejournal.com'.format(self.data['openid_lj_user'])



import random
import unittest

from ..strategy import BaseStrategy
from ..storage import UserMixin, NonceMixin, AssociationMixin, \
    CodeMixin, BaseStorage

from .models import User


NOT_IMPLEMENTED_MSG = 'Implement in subclass'


class BrokenUser(UserMixin):
    pass


class BrokenAssociation(AssociationMixin):
    pass


class BrokenNonce(NonceMixin):
    pass


class BrokenCode(CodeMixin):
    pass


class BrokenStrategy(BaseStrategy):
    pass


class BrokenStrategyWithSettings(BrokenStrategy):
    def get_setting(self, name):
        raise AttributeError()


class BrokenStorage(BaseStorage):
    pass


class BrokenUserTests(unittest.TestCase):

    def setUp(self):
        self.user = BrokenUser

    def tearDown(self):
        self.user = None

    def test_get_username(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.user.get_username(User('foobar'))

    def test_user_model(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.user.user_model()

    def test_username_max_length(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.user.username_max_length()

    def test_get_user(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.user.get_user(1)

    def test_get_social_auth(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.user.get_social_auth('foo', 1)

    def test_get_social_auth_for_user(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.user.get_social_auth_for_user(User('foobar'))

    def test_create_social_auth(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.user.create_social_auth(User('foobar'), 1, 'foo')

    def test_disconnect(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.user.disconnect(BrokenUser())


class BrokenAssociationTests(unittest.TestCase):
    def setUp(self):
        self.association = BrokenAssociation

    def tearDown(self):
        self.association = None

    def test_store(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.association.store('http://foobar.com', BrokenAssociation())

    def test_get(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.association.get()

    def test_remove(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.association.remove([1, 2, 3])


class BrokenNonceTests(unittest.TestCase):
    def setUp(self):
        self.nonce = BrokenNonce

    def tearDown(self):
        self.nonce = None

    def test_use(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.nonce.use('http://foobar.com', 1364951922, 'foobar123')


class BrokenCodeTest(unittest.TestCase):
    def setUp(self):
        self.code = BrokenCode

    def tearDown(self):
        self.code = None

    def test_get_code(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.code.get_code('foobar')


class BrokenStrategyTests(unittest.TestCase):
    def setUp(self):
        self.strategy = BrokenStrategy(storage=BrokenStorage)

    def tearDown(self):
        self.strategy = None

    def test_redirect(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.strategy.redirect('http://foobar.com')

    def test_get_setting(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.strategy.get_setting('foobar')

    def test_html(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.strategy.html('<p>foobar</p>')

    def test_request_data(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.strategy.request_data()

    def test_request_host(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.strategy.request_host()

    def test_session_get(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.strategy.session_get('foobar')

    def test_session_set(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.strategy.session_set('foobar', 123)

    def test_session_pop(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.strategy.session_pop('foobar')

    def test_build_absolute_uri(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.strategy.build_absolute_uri('/foobar')

    def test_render_html_with_tpl(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.strategy.render_html('foobar.html', context={})

    def test_render_html_with_html(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.strategy.render_html(html='<p>foobar</p>', context={})

    def test_render_html_with_none(self):
        with self.assertRaisesRegex(ValueError,
                                    'Missing template or html parameters'):
            self.strategy.render_html()

    def test_is_integrity_error(self):
        with self.assertRaisesRegex(NotImplementedError, NOT_IMPLEMENTED_MSG):
            self.strategy.storage.is_integrity_error(None)

    def test_random_string(self):
        self.assertTrue(isinstance(self.strategy.random_string(), str))

    def test_random_string_without_systemrandom(self):
        def SystemRandom():
            raise NotImplementedError()

        orig_random = getattr(random, 'SystemRandom', None)
        random.SystemRandom = SystemRandom

        strategy = BrokenStrategyWithSettings(storage=BrokenStorage)
        self.assertTrue(isinstance(strategy.random_string(), str))
        random.SystemRandom = orig_random



from .oauth import BaseOAuth2


class DigitalOceanOAuth(BaseOAuth2):
    """
    DigitalOcean OAuth authentication backend.

    Docs: https://developers.digitalocean.com/documentation/oauth/
    """
    name = 'digitalocean'
    AUTHORIZATION_URL = 'https://cloud.digitalocean.com/v1/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://cloud.digitalocean.com/v1/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ' '
    EXTRA_DATA = [
        ('expires_in', 'expires_in')
    ]

    def get_user_id(self, details, response):
        """Return user unique id provided by service"""
        return response['account'].get('uuid')

    def get_user_details(self, response):
        """Return user details from DigitalOcean account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('name') or '')

        return {'username': response['account'].get('email'),
                'email': response['account'].get('email'),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, token, *args, **kwargs):
        """Loads user data from service"""
        url = 'https://api.digitalocean.com/v2/account'
        auth_header = {'Authorization': 'Bearer %s' % token}
        try:
            return self.get_json(url, headers=auth_header)
        except ValueError:
            return None



import json

from .oauth import OAuth2Test


class EventbriteAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.eventbrite.EventbriteOAuth2'
    user_data_url = 'https://www.eventbriteapi.com/v3/users/me'
    expected_username = 'sean+awesome@eventbrite.com'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'first_name': 'sean',
        'last_name': 'rose',
        'name': 'sean rose',
        'access_token': 'YQEN5H2W2OTKQZWZMYER',
        'emails': [
            {
                'verified': True,
                'email': 'sean+awesome2@eventbrite.com',
                'primary': False,
            },
            {
                'verified': True,
                'email': 'sean+awesome@eventbrite.com',
                'primary': True,
            },
        ],
        'token_type': 'bearer',
        'image_id': None,
        'is_public': False,
        'id': '128559602587',
    })

    def test_login(self):
        self.do_login()



"""
Github Enterprise OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/github_enterprise.html
"""
from urllib.parse import urljoin

from ..utils import append_slash
from .github import GithubOAuth2, GithubOrganizationOAuth2, \
                    GithubTeamOAuth2


class GithubEnterpriseMixin:
    def api_url(self):
        return append_slash(self.setting('API_URL'))

    def authorization_url(self):
        return self._url('login/oauth/authorize')

    def access_token_url(self):
        return self._url('login/oauth/access_token')

    def _url(self, path):
        return urljoin(append_slash(self.setting('URL')), path)


class GithubEnterpriseOAuth2(GithubEnterpriseMixin, GithubOAuth2):
    """Github Enterprise OAuth authentication backend"""
    name = 'github-enterprise'


class GithubEnterpriseOrganizationOAuth2(GithubEnterpriseMixin,
                                         GithubOrganizationOAuth2):
    """Github Enterprise OAuth2 authentication backend for
    organizations"""
    name = 'github-enterprise-org'
    DEFAULT_SCOPE = ['read:org']


class GithubEnterpriseTeamOAuth2(GithubEnterpriseMixin, GithubTeamOAuth2):
    """Github Enterprise OAuth2 authentication backend for teams"""
    name = 'github-enterprise-team'
    DEFAULT_SCOPE = ['read:org']



"""
Mail.ru OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/mailru.html
"""
from hashlib import md5
from urllib.parse import unquote

from .oauth import BaseOAuth2


class MailruOAuth2(BaseOAuth2):
    """Mail.ru authentication backend"""
    name = 'mailru-oauth2'
    ID_KEY = 'uid'
    AUTHORIZATION_URL = 'https://connect.mail.ru/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://connect.mail.ru/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    EXTRA_DATA = [('refresh_token', 'refresh_token'),
                  ('expires_in', 'expires')]

    def get_user_details(self, response):
        """Return user details from Mail.ru request"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=unquote(response['first_name']),
            last_name=unquote(response['last_name'])
        )
        return {'username': unquote(response['nick']),
                'email': unquote(response['email']),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Return user data from Mail.ru REST API"""
        key, secret = self.get_key_and_secret()
        data = {'method': 'users.getInfo',
                'session_key': access_token,
                'app_id': key,
                'secure': '1'}
        param_list = sorted(list(item + '=' + data[item] for item in data))
        data['sig'] = md5(
            (''.join(param_list) + secret).encode('utf-8')
        ).hexdigest()
        return self.get_json('http://www.appsmail.ru/platform/api',
                             params=data)[0]


class MRGOAuth2(BaseOAuth2):

    name = 'mailru'
    ID_KEY = 'email'
    AUTHORIZATION_URL = 'https://oauth.mail.ru/login'
    ACCESS_TOKEN_URL = 'https://oauth.mail.ru/token'
    ACCESS_TOKEN_METHOD = 'POST'
    EXTRA_DATA = [('refresh_token', 'refresh_token'),
                  ('expires_in', 'expires')]
    REDIRECT_STATE = False

    def get_user_details(self, response):
        return {
            'gender': response.get('gender'),
            'fullname': response.get('name'),
            'username': response.get('name'),
            'first_name': response.get('first_name'),
            'last_name': response.get('last_name'),
            'locale': response.get('locale'),
            'email': response.get('email'),
            'address': response.get('address'),
            'birthday': response.get('birthday'),
            'image': response.get('image'),
        }

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json('https://oauth.mail.ru/userinfo', params={'access_token': access_token})



import json
from urllib.parse import urlencode

from .oauth import OAuth1Test


class TumblrOAuth1Test(OAuth1Test):
    backend_path = 'social_core.backends.tumblr.TumblrOAuth'
    user_data_url = 'http://api.tumblr.com/v2/user/info'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    request_token_body = urlencode({
        'oauth_token_secret': 'foobar-secret',
        'oauth_token': 'foobar',
        'oauth_callback_confirmed': 'true'
    })
    user_data_body = json.dumps({
        'meta': {
            'status': 200,
            'msg': 'OK'
        },
        'response': {
            'user': {
                'following': 1,
                'blogs': [{
                    'updated': 0,
                    'description': '',
                    'drafts': 0,
                    'title': 'Untitled',
                    'url': 'http://foobar.tumblr.com/',
                    'messages': 0,
                    'tweet': 'N',
                    'share_likes': True,
                    'posts': 0,
                    'primary': True,
                    'queue': 0,
                    'admin': True,
                    'followers': 0,
                    'ask': False,
                    'facebook': 'N',
                    'type': 'public',
                    'facebook_opengraph_enabled': 'N',
                    'name': 'foobar'
                }],
                'default_post_format': 'html',
                'name': 'foobar',
                'likes': 0
            }
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from httpretty import HTTPretty

from .oauth import OAuth2Test


class AtlassianOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.atlassian.AtlassianOAuth2'
    tenant_url = 'https://api.atlassian.com/oauth/token/accessible-resources'
    user_data_url = 'https://api.atlassian.com/ex/jira/FAKED_CLOUD_ID/rest/api/2/myself'
    expected_username = 'erlich'
    access_token_body = json.dumps({
        'access_token': 'aviato',
        'token_type': 'bearer'
    })
    tenant_data_body = json.dumps([
        {
            'id': 'FAKED_CLOUD_ID',
            'name': 'bachmanity.com',
            'avatarUrl': 'https://bachmanity.atlassian.net/avatars/240/site.png',
            'scopes': [
                'read:jira-user'
            ]
        }
    ])
    user_data_body = json.dumps({
        'self': 'http://bachmanity.atlassian.net/rest/api/3/user?username=erlich',
        'key': 'erlich',
        'accountId': '99:27935d01-92a7-4687-8272-a9b8d3b2ae2e',
        'name': 'erlich',
        'emailAddress': 'erlich@bachmanity.com',
        'avatarUrls': {
            '48x48': 'http://bachmanity.atlassian.net/secure/useravatar?size=large&ownerId=erlich',
            '24x24': 'http://bachmanity.atlassian.net/secure/useravatar?size=small&ownerId=erlich',
            '16x16': 'http://bachmanity.atlassian.net/secure/useravatar?size=xsmall&ownerId=erlich',
            '32x32': 'http://bachmanity.atlassian.net/secure/useravatar?size=medium&ownerId=erlich'
        },
        'displayName': 'Erlich Bachman',
        'active': True,
        'timeZone': 'Australia/Sydney',
        'groups': {
            'size': 3,
            'items': []
        },
        'applicationRoles': {
            'size': 1,
            'items': []
        }
    })

    def auth_handlers(self, start_url):
        target_url = super().auth_handlers(start_url)
        HTTPretty.register_uri(HTTPretty.GET,
                               self.tenant_url,
                               body=self.tenant_data_body,
                               content_type='application/json')
        return target_url

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Coding OAuth2 backend, docs at:
"""
from urllib.parse import urljoin

from .oauth import BaseOAuth2


class CodingOAuth2(BaseOAuth2):
    """Coding OAuth authentication backend"""

    name = 'coding'
    API_URL = 'https://coding.net/api/'
    AUTHORIZATION_URL = 'https://coding.net/oauth_authorize.html'
    ACCESS_TOKEN_URL = 'https://coding.net/api/oauth/access_token'
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ','
    DEFAULT_SCOPE = ['user']
    REDIRECT_STATE = False

    def api_url(self):
        return self.API_URL

    def get_user_details(self, response):
        """Return user details from Github account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('name')
        )
        return {'username': response.get('name'),
                'email': response.get('email') or '',
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        data = self._user_data(access_token)
        if data.get('code') != 0:
            # 获取失败
            pass
        return data.get('data')

    def _user_data(self, access_token, path=None):
        url = urljoin(
            self.api_url(),
            'account/current_user{0}'.format(path or '')
        )
        return self.get_json(url, params={'access_token': access_token})



"""
LinkedIn OAuth1 and OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/linkedin.html
"""
from social_core.exceptions import AuthCanceled
from .oauth import BaseOAuth2


class LinkedinOAuth2(BaseOAuth2):
    name = 'linkedin-oauth2'
    AUTHORIZATION_URL = \
        'https://www.linkedin.com/oauth/v2/authorization'
    ACCESS_TOKEN_URL = 'https://www.linkedin.com/oauth/v2/accessToken'
    USER_DETAILS_URL = \
        'https://api.linkedin.com/v2/me?projection=({projection})'
    USER_EMAILS_URL = 'https://api.linkedin.com/v2/emailAddress' \
                      '?q=members&projection=(elements*(handle~))'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    DEFAULT_SCOPE = ['r_liteprofile']
    EXTRA_DATA = [
        ('id', 'id'),
        ('expires_in', 'expires'),
        ('firstName', 'first_name'),
        ('lastName', 'last_name')
    ]

    def user_details_url(self):
        # use set() since LinkedIn fails when values are duplicated
        fields_selectors = list(set(['id', 'firstName', 'lastName'] +
                                    self.setting('FIELD_SELECTORS', [])))
        # user sort to ease the tests URL mocking
        fields_selectors.sort()
        fields_selectors = ','.join(fields_selectors)
        return self.USER_DETAILS_URL.format(projection=fields_selectors)

    def user_emails_url(self):
        return self.USER_EMAILS_URL

    def user_data(self, access_token, *args, **kwargs):
        response = self.get_json(
            self.user_details_url(),
            headers=self.user_data_headers(access_token)
        )

        if 'emailAddress' in set(self.setting('FIELD_SELECTORS', [])):
            emails = self.email_data(access_token, *args, **kwargs)
            if emails:
                response['emailAddress'] = emails[0]

        return response

    def email_data(self, access_token, *args, **kwargs):
        response = self.get_json(
            self.user_emails_url(),
            headers=self.user_data_headers(access_token)
        )
        email_addresses = []
        for element in response.get('elements', []):
            email_address = element.get('handle~', {}).get('emailAddress')
            email_addresses.append(email_address)
        return list(filter(None, email_addresses))

    def get_user_details(self, response):
        """Return user details from Linkedin account"""

        def get_localized_name(name):
            """
            FirstName & Last Name object
            {
                  'localized': {
                     'en_US': 'Smith'
                  },
                  'preferredLocale': {
                     'country': 'US',
                     'language': 'en'
                  }
            }
            :return the localizedName from the lastName object
            """
            locale = '{}_{}'.format(
                name['preferredLocale']['language'],
                name['preferredLocale']['country']
            )
            return name['localized'].get(locale, '')

        fullname, first_name, last_name = self.get_user_names(
            first_name=get_localized_name(response['firstName']),
            last_name=get_localized_name(response['lastName'])
        )
        email = response.get('emailAddress', '')
        return {'username': first_name + last_name,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name,
                'email': email}

    def user_data_headers(self, access_token):
        headers = {}
        lang = self.setting('FORCE_PROFILE_LANGUAGE')
        if lang:
            headers['Accept-Language'] = lang if lang is not True \
                else self.strategy.get_language()
        headers['Authorization'] = 'Bearer {access_token}'.format(
            access_token=access_token)
        return headers

    def request_access_token(self, *args, **kwargs):
        # LinkedIn expects a POST request with querystring parameters, despite
        # the spec http://tools.ietf.org/html/rfc6749#section-4.1.3
        kwargs['params'] = kwargs.pop('data')
        return super().request_access_token(*args, **kwargs)

    def process_error(self, data):
        super().process_error(data)
        if data.get('serviceErrorCode'):
            raise AuthCanceled(self, data.get('message') or data.get('status'))


class LinkedinMobileOAuth2(LinkedinOAuth2):
    name = 'linkedin-mobile-oauth2'

    def user_data_headers(self, access_token):
        headers = super().user_data_headers(access_token)
        headers['x-li-src'] = 'msdk'
        return headers



import hashlib

from ..utils import handle_http_errors
from .base import BaseAuth


class LastFmAuth(BaseAuth):
    """
    Last.Fm authentication backend. Requires two settings:
        SOCIAL_AUTH_LASTFM_KEY
        SOCIAL_AUTH_LASTFM_SECRET

    Don't forget to set the Last.fm callback to something sensible like
        http://your.site/lastfm/complete
    """
    name = 'lastfm'
    AUTH_URL = 'http://www.last.fm/api/auth/?api_key={api_key}'
    EXTRA_DATA = [
        ('key', 'session_key')
    ]

    def auth_url(self):
        return self.AUTH_URL.format(api_key=self.setting('KEY'))

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        key, secret = self.get_key_and_secret()
        token = self.data['token']

        signature = hashlib.md5(''.join(
            ('api_key', key, 'methodauth.getSession', 'token', token, secret)
        ).encode()).hexdigest()

        response = self.get_json('http://ws.audioscrobbler.com/2.0/', data={
            'method': 'auth.getSession',
            'api_key': key,
            'token': token,
            'api_sig': signature,
            'format': 'json'
        }, method='POST')

        kwargs.update({'response': response['session'], 'backend': self})
        return self.strategy.authenticate(*args, **kwargs)

    def get_user_id(self, details, response):
        """Return a unique ID for the current user, by default from server
        response."""
        return response.get('name')

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(response['name'])
        return {
            'username': response['name'],
            'email': '',
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name
        }



import json

from .oauth import OAuth2Test


class UberOAuth2Test(OAuth2Test):
    user_data_url = 'https://api.uber.com/v1/me'
    backend_path = 'social_core.backends.uber.UberOAuth2'
    expected_username = 'foo@bar.com'

    user_data_body = json.dumps({
        'first_name': 'Foo',
        'last_name': 'Bar',
        'email': 'foo@bar.com',
        'picture': 'https://',
        'promo_code': 'barfoo',
        'uuid': '91d81273-45c2-4b57-8124-d0165f8240c0'
    })

    access_token_body = json.dumps({
        'access_token': 'EE1IDxytP04tJ767GbjH7ED9PpGmYvL',
        'token_type': 'Bearer',
        'expires_in': 2592000,
        'refresh_token': 'Zx8fJ8qdSRRseIVlsGgtgQ4wnZBehr',
        'scope': 'profile history request'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Sign In With Apple authentication backend.

Docs:
    * https://developer.apple.com/documentation/signinwithapplerestapi
    * https://developer.apple.com/documentation/signinwithapplerestapi/tokenresponse

Settings:
    * `TEAM` - your team id;
    * `KEY` - your key id;
    * `CLIENT` - your client id;
    * `AUDIENCE` - a list of authorized client IDs, defaults to [CLIENT].
                   Use this if you need to accept both service and bundle id to
                   be able to login both via iOS and ie a web form.
    * `SECRET` - your secret key;
    * `SCOPE` (optional) - e.g. `['name', 'email']`;
    * `EMAIL_AS_USERNAME` - use apple email is username is set, use apple id
                            otherwise.
    * `AppleIdAuth.TOKEN_TTL_SEC` - time before JWT token expiration, seconds.
    * `SOCIAL_AUTH_APPLE_ID_INACTIVE_USER_LOGIN` - allow inactive users email to
                                                   login
"""

import json
import time

import jwt
from jwt.algorithms import RSAAlgorithm
from jwt.exceptions import PyJWTError

from social_core.backends.oauth import BaseOAuth2
from social_core.exceptions import AuthFailed


class AppleIdAuth(BaseOAuth2):
    name = 'apple-id'

    JWK_URL = 'https://appleid.apple.com/auth/keys'
    AUTHORIZATION_URL = 'https://appleid.apple.com/auth/authorize'
    ACCESS_TOKEN_URL = 'https://appleid.apple.com/auth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    RESPONSE_MODE = None

    ID_KEY = 'sub'
    TOKEN_KEY = 'id_token'
    STATE_PARAMETER = True
    REDIRECT_STATE = False
    SCOPE_SEPARATOR = '%20'

    TOKEN_AUDIENCE = 'https://appleid.apple.com'
    TOKEN_TTL_SEC = 6 * 30 * 24 * 60 * 60

    def get_audience(self):
        client_id = self.setting('CLIENT')
        return self.setting('AUDIENCE', default=[client_id])

    def auth_params(self, *args, **kwargs):
        """
        Apple requires to set `response_mode` to `form_post` if `scope`
        parameter is passed.
        """
        params = super().auth_params(*args, **kwargs)
        if self.RESPONSE_MODE:
            params['response_mode'] = self.RESPONSE_MODE
        elif self.get_scope():
            params['response_mode'] = 'form_post'
        return params

    def get_private_key(self):
        """
        Return contents of the private key file. Override this method to provide
        secret key from another source if needed.
        """
        return self.setting('SECRET')

    def generate_client_secret(self):
        now = int(time.time())
        client_id = self.setting('CLIENT')
        team_id = self.setting('TEAM')
        key_id = self.setting('KEY')
        private_key = self.get_private_key()

        headers = {'kid': key_id}
        payload = {
            'iss': team_id,
            'iat': now,
            'exp': now + self.TOKEN_TTL_SEC,
            'aud': self.TOKEN_AUDIENCE,
            'sub': client_id,
        }

        return jwt.encode(payload, key=private_key, algorithm='ES256',
                          headers=headers)

    def get_key_and_secret(self):
        client_id = self.setting('CLIENT')
        client_secret = self.generate_client_secret()
        return client_id, client_secret

    def get_apple_jwk(self, kid=None):
        """
        Return requested Apple public key or all available.
        """
        keys = self.get_json(url=self.JWK_URL).get('keys')

        if not isinstance(keys, list) or not keys:
            raise AuthFailed(self, 'Invalid jwk response')

        if kid:
            return json.dumps([key for key in keys if key['kid'] == kid][0])
        else:
            return (json.dumps(key) for key in keys)

    def decode_id_token(self, id_token):
        """
        Decode and validate JWT token from apple and return payload including
        user data.
        """
        if not id_token:
            raise AuthFailed(self, 'Missing id_token parameter')

        try:
            kid = jwt.get_unverified_header(id_token).get('kid')
            public_key = RSAAlgorithm.from_jwk(self.get_apple_jwk(kid))
            decoded = jwt.decode(
                id_token,
                key=public_key,
                audience=self.get_audience(),
                algorithms=['RS256'],
            )
        except PyJWTError:
            raise AuthFailed(self, 'Token validation failed')

        return decoded

    def get_user_details(self, response):
        name = json.loads(self.data.get('user', '{}')).get('name', {})
        fullname, first_name, last_name = self.get_user_names(
            fullname='',
            first_name=name.get('firstName', ''),
            last_name=name.get('lastName', '')
        )

        email = response.get('email', '')
        apple_id = response.get(self.ID_KEY, '')
        # prevent updating User with empty strings
        user_details = {
            'fullname': fullname or None,
            'first_name': first_name or None,
            'last_name': last_name or None,
            'email': email,
        }
        if email and self.setting('EMAIL_AS_USERNAME'):
            user_details['username'] = email
        if apple_id and not self.setting('EMAIL_AS_USERNAME'):
            user_details['username'] = apple_id

        return user_details

    def do_auth(self, access_token, *args, **kwargs):
        response = kwargs.pop('response', None) or {}
        jwt_string = response.get(self.TOKEN_KEY) or access_token

        if not jwt_string:
            raise AuthFailed(self, 'Missing id_token parameter')

        decoded_data = self.decode_id_token(jwt_string)
        return super().do_auth(access_token, response=decoded_data, *args, **kwargs)



import json
from urllib.parse import urlencode
from httpretty import HTTPretty

from .oauth import OAuth2Test
from social_core.backends.osso import OssoOAuth2

class OssoOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.osso.OssoOAuth2'
    user_data_url = 'https://demo.ossoapp.com/oauth/me'
    expected_username = 'user@example.com'
    access_token_body = json.dumps(
        {
            'access_token': '3633395cffe739bb87089235c152155ae73b6794f7af353b2aa189aeeacee1ec',
            'token_type': 'bearer',
            'expires_in': 600
        }
    )
    user_data_body = json.dumps(
        {
            'email': 'user@example.com',
            'id': 'f23611a5-2817-43e2-94b7-99b25235ad2d',
            'idp': 'Okta',
            'requested': {
                'email': None,
                'domain': 'example.com'
            }
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


import json
from .oauth import OAuth2Test


class ArcGISOAuth2Test(OAuth2Test):
    user_data_url = 'https://www.arcgis.com/sharing/rest/community/self'
    backend_path = 'social_core.backends.arcgis.ArcGISOAuth2'
    expected_username = 'gis@rocks.com'

    user_data_body = json.dumps({
        'first_name': 'Gis',
        'last_name': 'Rocks',
        'email': 'gis@rocks.com',
        'fullName': 'Gis Rocks',
        'username': 'gis@rocks.com'
    })

    access_token_body = json.dumps({
        'access_token': 'CM-gcB85taGhRmoI7l3PSGaXUNsaLkTg-dHH7XtA9Dnlin'
                        'PYKBBrIvFzhd1JtDhh7hEwSv_6eLLcLtUqe3gD6i1yaYYF'
                        'pUQJwy8KEujke5AE87tP9XIoMtp4_l320pUL',
        'expires_in': 86400
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Stripe OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/stripe.html
"""
from .oauth import BaseOAuth2


class StripeOAuth2(BaseOAuth2):
    """Stripe OAuth2 authentication backend"""
    name = 'stripe'
    ID_KEY = 'stripe_user_id'
    AUTHORIZATION_URL = 'https://connect.stripe.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://connect.stripe.com/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('stripe_publishable_key', 'stripe_publishable_key'),
        ('access_token', 'access_token'),
        ('livemode', 'livemode'),
        ('token_type', 'token_type'),
        ('refresh_token', 'refresh_token'),
        ('stripe_user_id', 'stripe_user_id'),
    ]

    def get_user_details(self, response):
        """Return user details from Stripe account"""
        return {'username': response.get('stripe_user_id'),
                'email': ''}

    def auth_complete_params(self, state=None):
        client_id, client_secret = self.get_key_and_secret()
        return {
            'grant_type': 'authorization_code',
            'client_id': client_id,
            'scope': self.SCOPE_SEPARATOR.join(self.get_scope()),
            'code': self.data['code']
        }

    def auth_headers(self):
        client_id, client_secret = self.get_key_and_secret()
        return {'Accept': 'application/json',
                'Authorization': 'Bearer {0}'.format(client_secret)}

    def refresh_token_params(self, refresh_token, *args, **kwargs):
        return {'refresh_token': refresh_token,
                'grant_type': 'refresh_token'}



"""
    ORCID OAuth2 Application backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/orcid.html
"""
from .oauth import BaseOAuth2

class ORCIDOAuth2(BaseOAuth2):
    """ORCID OAuth2 authentication backend"""
    name = 'orcid'
    ID_KEY = 'orcid'
    AUTHORIZATION_URL = 'https://orcid.org/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://orcid.org/oauth/token'
    USER_ID_URL = 'https://orcid.org/oauth/userinfo'
    USER_DATA_URL = 'https://pub.orcid.org/v2.0/{}'
    DEFAULT_SCOPE = ['/authenticate']
    ACCESS_TOKEN_METHOD = 'POST'
    EXTRA_DATA = [
        ('orcid', 'id'),
        ('expires_in', 'expires'),
        ('refresh_token', 'refresh_token')
    ]

    def auth_params(self, state=None):
        params = super().auth_params(state)
        return params

    def get_user_details(self, response):
        """Return user details from ORCID account"""

        # response data will be of the following format:
        # {
        #     'orcid-identifier': {
        #         'uri': 'http://orcid.org/0000-0002-2601-8132',
        #         'path': '0000-0002-2601-8132',
        #         'host': 'orcid.org'
        #     },
        #     'person': {
        #         'last-modified-date': None,
        #         'name': {
        #             'created-date': {
        #                 'value': 1578249746904
        #             },
        #             'last-modified-date': {
        #                 'value': 1578249746904
        #             },
        #             'given-names': {
        #                 'value': 'Janani Kantharooban'
        #             },
        #             'family-name': {
        #                 'value': 'Umachanger'
        #             },
        #             'credit-name': None,
        #             'source': None,
        #             'visibility': 'PUBLIC',
        #             'path': '0000-0002-2601-8132'
        #         },
        #     }
        # }
        orcid_identifier = response.get('orcid-identifier')

        fullname = first_name = last_name = email = username = ''

        person = response.get('person')

        # Although we're checking here, the response will always have the orcid-identifier key:
        if orcid_identifier:
            username = orcid_identifier['path']

        if person:
            name = person.get('name')

            fullname = name

            if name:
                first_name = name.get('given-names', {}).get('value', '')
                last_name = name.get('family-name', {}).get('value', '')

            emails = person.get('emails')
            if emails:
                emails_list = emails.get('email')
                if emails_list and len(emails_list) > 0:
                    email = emails_list[0].get('email', '')

                    if len(emails_list) > 1:
                        for email_dict in emails_list:
                            if email_dict.get('primary'):
                                email = email_dict['email']
                                break
                    else:
                        email = emails_list[0].get('email', '')

        return {
            'username': username,
            'email': email,
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        params = self.setting('PROFILE_EXTRA_PARAMS', {})
        params['access_token'] = access_token

        # Reference Docs: ORCID Auth Flow: https://github.com/ORCID/ORCID-Source/blob/master/orcid-web/ORCID_AUTH_WITH_OPENID_CONNECT.md#other-endpoints
        # Sample headers: -H "Accept: application/json" -H "Authorization: Bearer <access_token>"
        # This will respond with a json document like this:
        # {
        #     "sub":"0000-0002-2601-8132",
        #     "name":"Credit Name",
        #     "family_name":"Jones",
        #     "given_name":"Tom"
        # }
        try:
            response = self.get_json(
                self.USER_ID_URL,
                headers={
                    'Content-Type': 'application/json',
                    'Authorization': 'Bearer {}'.format(str(access_token))
                },
            )

            # Update Jan 28 2021: Now we definitely have an ORCID id of format "0000-0000-0000-0000"
            orcid = response['sub']
        except Exception as ex:
            pass

        # We can now attempt to access the ORCID public API with the Orcid:
        try:
            return self.get_json(
                    self.USER_DATA_URL.format(orcid),
                    headers={
                        'Content-Type': 'application/json'
                    },
                    params=params
                )
        except Exception as ex:
            return None


class ORCIDOAuth2Sandbox(ORCIDOAuth2):
    """ORCID OAuth2 Sandbox authentication backend"""
    name = 'orcid-sandbox'
    AUTHORIZATION_URL = 'https://sandbox.orcid.org/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://sandbox.orcid.org/oauth/token'
    USER_ID_URL = 'https://sandbox.orcid.org/oauth/userinfo'
    USER_DATA_URL = 'https://pub.sandbox.orcid.org/v2.0/{}'


class ORCIDMemberOAuth2(ORCIDOAuth2):
    """ORCID OAuth2 authentication backend that uses ORCID Member API"""
    USER_DATA_URL = 'https://api.orcid.org/v2.0/{}'
    DEFAULT_SCOPE = ['/authenticate', '/read-limited']


class ORCIDMemberOAuth2Sandbox(ORCIDOAuth2Sandbox):
    """ORCID OAuth2 Sandbox authentication backend that uses ORCID Member Sandbox API"""
    USER_DATA_URL = 'https://api.sandbox.orcid.org/v2.0/{}'
    DEFAULT_SCOPE = ['/authenticate', '/read-limited']



import json

from .oauth import OAuth2Test


class GitLabOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.gitlab.GitLabOAuth2'
    user_data_url = 'https://gitlab.com/api/v4/user'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer',
        'expires_in': 7200,
        'refresh_token': 'barfoo'
    })
    user_data_body = json.dumps({
        'two_factor_enabled': False,
        'can_create_project': True,
        'confirmed_at': '2016-12-28T12:26:19.256Z',
        'twitter': '',
        'linkedin': '',
        'color_scheme_id': 1,
        'web_url': 'https://gitlab.com/foobar',
        'skype': '',
        'identities': [],
        'id': 123456,
        'projects_limit': 100000,
        'current_sign_in_at': '2016-12-28T12:26:19.795Z',
        'state': 'active',
        'location': None,
        'email': 'foobar@example.com',
        'website_url': '',
        'username': 'foobar',
        'bio': None,
        'last_sign_in_at': '2016-12-28T12:26:19.795Z',
        'is_admin': False,
        'external': False,
        'organization': None,
        'name': 'Foo Bar',
        'can_create_group': True,
        'created_at': '2016-12-28T12:26:19.638Z',
        'avatar_url': 'https://secure.gravatar.com/avatar/94d093eda664addd6e450d7e9881bcae?s=32&d=identicon',
        'theme_id': 2
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class GitLabCustomDomainOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.gitlab.GitLabOAuth2'
    user_data_url = 'https://example.com/api/v4/user'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer',
        'expires_in': 7200,
        'refresh_token': 'barfoo'
    })
    user_data_body = json.dumps({
        'two_factor_enabled': False,
        'can_create_project': True,
        'confirmed_at': '2016-12-28T12:26:19.256Z',
        'twitter': '',
        'linkedin': '',
        'color_scheme_id': 1,
        'web_url': 'https://example.com/foobar',
        'skype': '',
        'identities': [],
        'id': 123456,
        'projects_limit': 100000,
        'current_sign_in_at': '2016-12-28T12:26:19.795Z',
        'state': 'active',
        'location': None,
        'email': 'foobar@example.com',
        'website_url': '',
        'username': 'foobar',
        'bio': None,
        'last_sign_in_at': '2016-12-28T12:26:19.795Z',
        'is_admin': False,
        'external': False,
        'organization': None,
        'name': 'Foo Bar',
        'can_create_group': True,
        'created_at': '2016-12-28T12:26:19.638Z',
        'avatar_url': 'https://secure.gravatar.com/avatar/94d093eda664addd6e450d7e9881bcae?s=32&d=identicon',
        'theme_id': 2
    })

    def test_login(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITLAB_API_URL': 'https://example.com'
        })
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITLAB_API_URL': 'https://example.com'
        })
        self.do_partial_pipeline()



from .oauth import BaseOAuth2
from urllib.parse import urlencode

class OssoOAuth2(BaseOAuth2):
    """Osso OAuth authentication backend"""
    name = 'osso'
    REDIRECT_STATE = False
    STATE_PARAMETER = True
    AUTHORIZATION_URL = '{osso_base_url}/oauth/authorize'
    ACCESS_TOKEN_URL = '{osso_base_url}/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'

    @property
    def osso_base_url(self):
        return self.setting('OSSO_BASE_URL', 'https://demo.ossoapp.com')

    def authorization_url(self):
        return self.AUTHORIZATION_URL.format(osso_base_url=self.osso_base_url)

    def access_token_url(self):
        return self.ACCESS_TOKEN_URL.format(osso_base_url=self.osso_base_url)

    def auth_params(self, state=None):
        client_id, _client_secret = self.get_key_and_secret()
        params = {
            'client_id': client_id,
            'redirect_uri': self.get_redirect_uri(state)
        }
        if self.data.get('email'):
            params['email'] = self.data.get('email')
        if self.data.get('domain') and not self.data.get('email'):
            params['domain'] = self.data.get('domain')
        if self.STATE_PARAMETER and state:
            params['state'] = state
        if self.RESPONSE_TYPE:
            params['response_type'] = self.RESPONSE_TYPE
        return params


    def get_user_details(self, response):
        """Return user details from Osso"""
        return {'username': response.get('email'),
                'email': response.get('email')}

    def user_data(self, access_token, *args, **kwargs):
        """Loads normalized user profile from Osso"""
        url = '{osso_base_url}/oauth/me?'.format(osso_base_url=self.osso_base_url) + urlencode({
            'access_token': access_token
        })
        return self.get_json(url)


import json

from httpretty import HTTPretty

from ...exceptions import AuthFailed

from .oauth import OAuth2Test


class GithubEnterpriseOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.github_enterprise.GithubEnterpriseOAuth2'
    user_data_url = 'https://www.example.com/api/v3/user'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'login': 'foobar',
        'id': 1,
        'avatar_url': 'https://www.example.com/images/error/foobar_happy.gif',
        'gravatar_id': 'somehexcode',
        'url': 'https://www.example.com/api/v3/users/foobar',
        'name': 'monalisa foobar',
        'company': 'GitHub',
        'blog': 'https://www.example.com/blog',
        'location': 'San Francisco',
        'email': 'foo@bar.com',
        'hireable': False,
        'bio': 'There once was...',
        'public_repos': 2,
        'public_gists': 1,
        'followers': 20,
        'following': 0,
        'html_url': 'https://www.example.com/foobar',
        'created_at': '2008-01-14T04:33:35Z',
        'type': 'User',
        'total_private_repos': 100,
        'owned_private_repos': 100,
        'private_gists': 81,
        'disk_usage': 10000,
        'collaborators': 8,
        'plan': {
            'name': 'Medium',
            'space': 400,
            'collaborators': 10,
            'private_repos': 20
        }
    })

    def test_login(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_URL': 'https://www.example.com'})
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_API_URL': 'https://www.example.com/api/v3'})
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_URL': 'https://www.example.com'})
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_API_URL': 'https://www.example.com/api/v3'})
        self.do_partial_pipeline()


class GithubEnterpriseOAuth2NoEmailTest(GithubEnterpriseOAuth2Test):
    user_data_body = json.dumps({
        'login': 'foobar',
        'id': 1,
        'avatar_url': 'https://www.example.com/images/error/foobar_happy.gif',
        'gravatar_id': 'somehexcode',
        'url': 'https://www.example.com/api/v3/users/foobar',
        'name': 'monalisa foobar',
        'company': 'GitHub',
        'blog': 'https://www.example.com/blog',
        'location': 'San Francisco',
        'email': '',
        'hireable': False,
        'bio': 'There once was...',
        'public_repos': 2,
        'public_gists': 1,
        'followers': 20,
        'following': 0,
        'html_url': 'https://www.example.com/foobar',
        'created_at': '2008-01-14T04:33:35Z',
        'type': 'User',
        'total_private_repos': 100,
        'owned_private_repos': 100,
        'private_gists': 81,
        'disk_usage': 10000,
        'collaborators': 8,
        'plan': {
            'name': 'Medium',
            'space': 400,
            'collaborators': 10,
            'private_repos': 20
        }
    })

    def test_login(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_URL': 'https://www.example.com'})
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_API_URL': 'https://www.example.com/api/v3'})
        url = 'https://www.example.com/api/v3/user/emails'
        HTTPretty.register_uri(HTTPretty.GET, url, status=200,
                               body=json.dumps(['foo@bar.com']),
                               content_type='application/json')
        self.do_login()

    def test_login_next_format(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_URL': 'https://www.example.com'})
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_API_URL': 'https://www.example.com/api/v3'})
        url = 'https://www.example.com/api/v3/user/emails'
        HTTPretty.register_uri(HTTPretty.GET, url, status=200,
                               body=json.dumps([{'email': 'foo@bar.com'}]),
                               content_type='application/json')
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_URL': 'https://www.example.com'})
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_API_URL': 'https://www.example.com/api/v3'})
        self.do_partial_pipeline()


class GithubEnterpriseOrganizationOAuth2Test(GithubEnterpriseOAuth2Test):
    backend_path = 'social_core.backends.github_enterprise.GithubEnterpriseOrganizationOAuth2'

    def auth_handlers(self, start_url):
        url = 'https://www.example.com/api/v3/orgs/foobar/members/foobar'
        HTTPretty.register_uri(HTTPretty.GET, url, status=204, body='')
        return super().auth_handlers(start_url)

    def test_login(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_URL': 'https://www.example.com'})
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_API_URL': 'https://www.example.com/api/v3'})
        self.strategy.set_settings({'SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_NAME': 'foobar'})
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_URL': 'https://www.example.com'})
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_API_URL': 'https://www.example.com/api/v3'})
        self.strategy.set_settings({'SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_NAME': 'foobar'})
        self.do_partial_pipeline()


class GithubEnterpriseOrganizationOAuth2FailTest(GithubEnterpriseOAuth2Test):
    backend_path = 'social_core.backends.github_enterprise.GithubEnterpriseOrganizationOAuth2'

    def auth_handlers(self, start_url):
        url = 'https://www.example.com/api/v3/orgs/foobar/members/foobar'
        HTTPretty.register_uri(HTTPretty.GET, url, status=404,
                               body='{"message": "Not Found"}',
                               content_type='application/json')
        return super().auth_handlers(start_url)

    def test_login(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_URL': 'https://www.example.com'})
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_API_URL': 'https://www.example.com/api/v3'})
        self.strategy.set_settings({'SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_NAME': 'foobar'})
        with self.assertRaises(AuthFailed):
            self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_URL': 'https://www.example.com'})
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_API_URL': 'https://www.example.com/api/v3'})
        self.strategy.set_settings({'SOCIAL_AUTH_GITHUB_ENTERPRISE_ORG_NAME': 'foobar'})
        with self.assertRaises(AuthFailed):
            self.do_partial_pipeline()


class GithubEnterpriseTeamOAuth2Test(GithubEnterpriseOAuth2Test):
    backend_path = 'social_core.backends.github_enterprise.GithubEnterpriseTeamOAuth2'

    def auth_handlers(self, start_url):
        url = 'https://www.example.com/api/v3/teams/123/members/foobar'
        HTTPretty.register_uri(HTTPretty.GET, url, status=204, body='')
        return super().auth_handlers(start_url)

    def test_login(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_URL': 'https://www.example.com'})
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_API_URL': 'https://www.example.com/api/v3'})
        self.strategy.set_settings({'SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_ID': '123'})
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_URL': 'https://www.example.com'})
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_API_URL': 'https://www.example.com/api/v3'})
        self.strategy.set_settings({'SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_ID': '123'})
        self.do_partial_pipeline()


class GithubEnterpriseTeamOAuth2FailTest(GithubEnterpriseOAuth2Test):
    backend_path = 'social_core.backends.github_enterprise.GithubEnterpriseTeamOAuth2'

    def auth_handlers(self, start_url):
        url = 'https://www.example.com/api/v3/teams/123/members/foobar'
        HTTPretty.register_uri(HTTPretty.GET, url, status=404,
                               body='{"message": "Not Found"}',
                               content_type='application/json')
        return super().auth_handlers(start_url)

    def test_login(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_URL': 'https://www.example.com'})
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_API_URL': 'https://www.example.com/api/v3'})
        self.strategy.set_settings({'SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_ID': '123'})
        with self.assertRaises(AuthFailed):
            self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_URL': 'https://www.example.com'})
        self.strategy.set_settings({
            'SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_API_URL': 'https://www.example.com/api/v3'})
        self.strategy.set_settings({'SOCIAL_AUTH_GITHUB_ENTERPRISE_TEAM_ID': '123'})
        with self.assertRaises(AuthFailed):
            self.do_partial_pipeline()



"""Tests for NGP VAN ActionID Backend"""
import datetime
from urllib.parse import urlencode

from httpretty import HTTPretty

from .open_id import OpenIdTest


JANRAIN_NONCE = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')


class NGPVANActionIDOpenIDTest(OpenIdTest):
    """Test the NGP VAN ActionID OpenID 1.1 Backend"""
    backend_path = 'social_core.backends.ngpvan.ActionIDOpenID'
    expected_username = 'testuser@user.local'
    discovery_body = ' '.join([
        '<?xml version="1.0" encoding="UTF-8"?>',
        '<xrds:XRDS',
        'xmlns:xrds="xri://$xrds"',
        'xmlns:openid="http://openid.net/xmlns/1.0"',
        'xmlns="xri://$xrd*($v*2.0)">',
        '<XRD>',
        '<Service priority="10">',
        '<Type>http://specs.openid.net/auth/2.0/signon</Type>',
        '<Type>http://openid.net/extensions/sreg/1.1</Type>',
        '<Type>http://axschema.org/contact/email</Type>',
        '<URI>https://accounts.ngpvan.com/OpenId/Provider</URI>',
        '</Service>',
        '<Service priority="20">',
        '<Type>http://openid.net/signon/1.0</Type>',
        '<Type>http://openid.net/extensions/sreg/1.1</Type>',
        '<Type>http://axschema.org/contact/email</Type>',
        '<URI>https://accounts.ngpvan.com/OpenId/Provider</URI>',
        '</Service>',
        '</XRD>',
        '</xrds:XRDS>'
    ])
    server_response = urlencode({
        'openid.claimed_id': 'https://accounts.ngpvan.com/user/abcd123',
        'openid.identity': 'https://accounts.ngpvan.com/user/abcd123',
        'openid.sig': 'Midw8F/rCDwW7vMz3y+vK6rjz6s=',
        'openid.signed': 'claimed_id,identity,assoc_handle,op_endpoint,return_'
                         'to,response_nonce,ns.alias3,alias3.mode,alias3.type.'
                         'alias1,alias3.value.alias1,alias3.type.alias2,alias3'
                         '.value.alias2,alias3.type.alias3,alias3.value.alias3'
                         ',alias3.type.alias4,alias3.value.alias4,alias3.type.'
                         'alias5,alias3.value.alias5,alias3.type.alias6,alias3'
                         '.value.alias6,alias3.type.alias7,alias3.value.alias7'
                         ',alias3.type.alias8,alias3.value.alias8,ns.sreg,sreg'
                         '.fullname',
        'openid.assoc_handle': '{635790678917902781}{GdSyFA==}{20}',
        'openid.op_endpoint': 'https://accounts.ngpvan.com/OpenId/Provider',
        'openid.return_to': 'http://myapp.com/complete/actionid-openid/',
        'openid.response_nonce': JANRAIN_NONCE + 'MMgBGEre',
        'openid.mode': 'id_res',
        'openid.ns': 'http://specs.openid.net/auth/2.0',
        'openid.ns.alias3': 'http://openid.net/srv/ax/1.0',
        'openid.alias3.mode': 'fetch_response',
        'openid.alias3.type.alias1': 'http://openid.net/schema/contact/phone/b'
                                     'usiness',
        'openid.alias3.value.alias1': '+12015555555',
        'openid.alias3.type.alias2': 'http://openid.net/schema/contact/interne'
                                     't/email',
        'openid.alias3.value.alias2': 'testuser@user.local',
        'openid.alias3.type.alias3': 'http://openid.net/schema/namePerson/firs'
                                     't',
        'openid.alias3.value.alias3': 'John',
        'openid.alias3.type.alias4': 'http://openid.net/schema/namePerson/las'
                                     't',
        'openid.alias3.value.alias4': 'Smith',
        'openid.alias3.type.alias5': 'http://axschema.org/namePerson/first',
        'openid.alias3.value.alias5': 'John',
        'openid.alias3.type.alias6': 'http://axschema.org/namePerson/last',
        'openid.alias3.value.alias6': 'Smith',
        'openid.alias3.type.alias7': 'http://axschema.org/namePerson',
        'openid.alias3.value.alias7': 'John Smith',
        'openid.alias3.type.alias8': 'http://openid.net/schema/namePerson',
        'openid.alias3.value.alias8': 'John Smith',
        'openid.ns.sreg': 'http://openid.net/extensions/sreg/1.1',
        'openid.sreg.fullname': 'John Smith',
    })

    def setUp(self):
        """Setup the test"""
        super().setUp()

        # Mock out the NGP VAN endpoints
        HTTPretty.register_uri(
            HTTPretty.POST,
            'https://accounts.ngpvan.com/Home/Xrds',
            status=200,
            body=self.discovery_body
        )
        HTTPretty.register_uri(
            HTTPretty.GET,
            'https://accounts.ngpvan.com/user/abcd123',
            status=200,
            body=self.discovery_body
        )
        HTTPretty.register_uri(
            HTTPretty.GET,
            'https://accounts.ngpvan.com/OpenId/Provider',
            status=200,
            body=self.discovery_body
        )

    def test_login(self):
        """Test the login flow using python-social-auth's built in test"""
        self.do_login()

    def test_partial_pipeline(self):
        """Test the partial flow using python-social-auth's built in test"""
        self.do_partial_pipeline()

    def test_get_ax_attributes(self):
        """Test that the AX attributes that NGP VAN responds with are present"""
        records = self.backend.get_ax_attributes()

        self.assertEqual(records, [
            ('http://openid.net/schema/contact/internet/email', 'email'),
            ('http://openid.net/schema/contact/phone/business', 'phone'),
            ('http://openid.net/schema/namePerson/first', 'first_name'),
            ('http://openid.net/schema/namePerson/last', 'last_name'),
            ('http://openid.net/schema/namePerson', 'fullname'),
        ])

    def test_setup_request(self):
        """Test the setup_request functionality in the NGP VAN backend"""
        # We can grab the requested attributes by grabbing the HTML of the
        # OpenID auth form and pulling out the hidden fields
        _, inputs = self.get_form_data(self.backend.auth_html())

        # Confirm that the only required attribute is email
        self.assertEqual(inputs['openid.ax.required'], 'ngpvanemail')

        # Confirm that the 3 optional attributes are requested "if available"
        self.assertIn('ngpvanphone', inputs['openid.ax.if_available'])
        self.assertIn('ngpvanfirstname', inputs['openid.ax.if_available'])
        self.assertIn('ngpvanlastname', inputs['openid.ax.if_available'])

        # Verify the individual attribute properties
        self.assertEqual(
            inputs['openid.ax.type.ngpvanemail'],
            'http://openid.net/schema/contact/internet/email'
        )
        self.assertEqual(
            inputs['openid.ax.type.ngpvanfirstname'],
            'http://openid.net/schema/namePerson/first'
        )
        self.assertEqual(
            inputs['openid.ax.type.ngpvanlastname'],
            'http://openid.net/schema/namePerson/last'
        )
        self.assertEqual(
            inputs['openid.ax.type.ngpvanphone'],
            'http://openid.net/schema/contact/phone/business'
        )

    def test_user_data(self):
        """Ensure that the correct user data is being passed to create_user"""
        self.strategy.set_settings({
            'USER_FIELDS': [
                'email',
                'first_name',
                'last_name',
                'username',
                'phone',
                'fullname'
            ]
        })
        user = self.do_start()
        self.assertEqual(user.username, u'testuser@user.local')
        self.assertEqual(user.email, u'testuser@user.local')
        self.assertEqual(user.extra_user_fields['phone'], u'+12015555555')
        self.assertEqual(user.extra_user_fields['first_name'], u'John')
        self.assertEqual(user.extra_user_fields['last_name'], u'Smith')
        self.assertEqual(user.extra_user_fields['fullname'], u'John Smith')

    def test_extra_data_phone(self):
        """Confirm that you can get a phone number via the relevant setting"""
        self.strategy.set_settings({
            'SOCIAL_AUTH_ACTIONID_OPENID_AX_EXTRA_DATA': [
                ('http://openid.net/schema/contact/phone/business', 'phone')
            ]
        })
        user = self.do_start()
        self.assertEqual(user.social_user.extra_data['phone'], u'+12015555555')

    def test_association_uid(self):
        """Test that the correct association uid is stored in the database"""
        user = self.do_start()
        self.assertEqual(
            user.social_user.uid,
            'https://accounts.ngpvan.com/user/abcd123'
        )



"""
ExactTarget OAuth support.
Support Authentication from IMH using JWT token and pre-shared key.
Requires package pyjwt
"""
from datetime import timedelta, datetime

import jwt

from ..exceptions import AuthFailed, AuthCanceled
from .oauth import BaseOAuth2


class ExactTargetOAuth2(BaseOAuth2):
    name = 'exacttarget'

    def get_user_details(self, response):
        """Use the email address of the user, suffixed by _et"""
        user = response.get('token', {})\
                       .get('request', {})\
                       .get('user', {})
        if 'email' in user:
            user['username'] = user['email']
        return user

    def get_user_id(self, details, response):
        """
        Create a user ID from the ET user ID. Uses details rather than the
        default response, as only the token is available in response. details
        is much richer:
            {
                'expiresIn': 1200,
                'username': 'example@example.com',
                'refreshToken': '1234567890abcdef',
                'internalOauthToken': 'jwttoken.......',
                'oauthToken': 'yetanothertoken',
                'id': 123456,
                'culture': 'en-US',
                'timezone': {
                    'shortName': 'CST',
                    'offset': -6.0,
                    'dst': False,
                    'longName': '(GMT-06:00) Central Time (No Daylight Saving)'
                },
                'email': 'example@example.com'
            }
        """
        return '{0}'.format(details.get('id'))

    def uses_redirect(self):
        return False

    def auth_url(self):
        return None

    def process_error(self, data):
        if data.get('error'):
            error = self.data.get('error_description') or self.data['error']
            raise AuthFailed(self, error)

    def do_auth(self, token, *args, **kwargs):
        dummy, secret = self.get_key_and_secret()
        try:  # Decode the token, using the Application Signature from settings
            decoded = jwt.decode(token, secret, algorithms=['HS256'])
        except jwt.DecodeError:  # Wrong signature, fail authentication
            raise AuthCanceled(self)
        kwargs.update({'response': {'token': decoded}, 'backend': self})
        return self.strategy.authenticate(*args, **kwargs)

    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        token = self.data.get('jwt', {})
        if not token:
            raise AuthFailed(self, 'Authentication Failed')
        return self.do_auth(token, *args, **kwargs)

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        """Load extra details from the JWT token"""
        data = {
            'id': details.get('id'),
            'email': details.get('email'),
            # OAuth token, for use with legacy SOAP API calls:
            #   http://bit.ly/13pRHfo
            'internalOauthToken': details.get('internalOauthToken'),
            # Token for use with the Application ClientID for the FUEL API
            'oauthToken': details.get('oauthToken'),
            # If the token has expired, use the FUEL API to get a new token see
            # http://bit.ly/10v1K5l and http://bit.ly/11IbI6F - set legacy=1
            'refreshToken': details.get('refreshToken'),
        }

        # The expiresIn value determines how long the tokens are valid for.
        # Take a bit off, then convert to an int timestamp
        expiresSeconds = details.get('expiresIn', 0) - 30
        expires = datetime.utcnow() + timedelta(seconds=expiresSeconds)
        data['expires'] = (expires - datetime(1970, 1, 1)).total_seconds()

        if response.get('token'):
            token = response['token']
            org = token.get('request', {}).get('organization')
            if org:
                data['stack'] = org.get('stackKey')
                data['enterpriseId'] = org.get('enterpriseId')
        return data



from .oauth import BaseOAuth2


class UniverseOAuth2(BaseOAuth2):
    """Universe Ticketing OAuth2 authentication backend"""
    name = 'universe'
    AUTHORIZATION_URL = 'https://www.universe.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://www.universe.com/oauth/token'
    BASE_API_URL = 'https://www.universe.com/api'
    USER_INFO_URL = BASE_API_URL + '/v2/current_user'
    ACCESS_TOKEN_METHOD = 'POST'
    STATE_PARAMETER = True
    REDIRECT_STATE = True
    EXTRA_DATA = [
        ('id', 'id'),
        ('slug', 'slug'),
        ('created_at', 'created_at'),
        ('updated_at', 'updated_at'),
    ]

    def get_user_id(self, details, response):
        return response['current_user'][self.ID_KEY]

    def get_user_details(self, response):
        """Return user details from a Universe account"""
        # Start with the user data as it was returned
        user_details = response['current_user']
        user_details['username'] = user_details['email']
        return user_details

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(self.USER_INFO_URL, headers={
            'Authorization': 'Bearer {}'.format(access_token)
        })



import json
from urllib.parse import urlencode

from .oauth import OAuth1Test


class TwitterOAuth1Test(OAuth1Test):
    backend_path = 'social_core.backends.twitter.TwitterOAuth'
    user_data_url = 'https://api.twitter.com/1.1/account/' \
        'verify_credentials.json'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    request_token_body = urlencode({
        'oauth_token_secret': 'foobar-secret',
        'oauth_token': 'foobar',
        'oauth_callback_confirmed': 'true'
    })
    user_data_body = json.dumps({
        'follow_request_sent': False,
        'profile_use_background_image': True,
        'id': 10101010,
        'description': 'Foo bar baz qux',
        'verified': False,
        'entities': {
            'description': {
                'urls': []
            }
        },
        'profile_image_url_https': 'https://twimg0-a.akamaihd.net/'
                                   'profile_images/532018826/'
                                   'n587119531_1939735_9305_normal.jpg',
        'profile_sidebar_fill_color': '252429',
        'profile_text_color': '666666',
        'followers_count': 77,
        'profile_sidebar_border_color': '181A1E',
        'location': 'Fooland',
        'default_profile_image': False,
        'listed_count': 4,
        'status': {
            'favorited': False,
            'contributors': None,
            'retweeted_status': {
                'favorited': False,
                'contributors': None,
                'truncated': False,
                'source': 'web',
                'text': '"Foo foo foo foo',
                'created_at': 'Fri Dec 21 18:12:00 +0000 2012',
                'retweeted': True,
                'in_reply_to_status_id': None,
                'coordinates': None,
                'id': 101010101010101010,
                'entities': {
                    'user_mentions': [],
                    'hashtags': [],
                    'urls': []
                },
                'in_reply_to_status_id_str': None,
                'place': None,
                'id_str': '101010101010101010',
                'in_reply_to_screen_name': None,
                'retweet_count': 8,
                'geo': None,
                'in_reply_to_user_id_str': None,
                'in_reply_to_user_id': None
            },
            'truncated': False,
            'source': 'web',
            'text': 'RT @foo: "Foo foo foo foo',
            'created_at': 'Fri Dec 21 18:24:10 +0000 2012',
            'retweeted': True,
            'in_reply_to_status_id': None,
            'coordinates': None,
            'id': 101010101010101010,
            'entities': {
                'user_mentions': [{
                    'indices': [3, 10],
                    'id': 10101010,
                    'screen_name': 'foo',
                    'id_str': '10101010',
                    'name': 'Foo'
                }],
                'hashtags': [],
                'urls': []
            },
            'in_reply_to_status_id_str': None,
            'place': None,
            'id_str': '101010101010101010',
            'in_reply_to_screen_name': None,
            'retweet_count': 8,
            'geo': None,
            'in_reply_to_user_id_str': None,
            'in_reply_to_user_id': None
        },
        'utc_offset': -10800,
        'statuses_count': 191,
        'profile_background_color': '1A1B1F',
        'friends_count': 151,
        'profile_background_image_url_https': 'https://twimg0-a.akamaihd.net/'
                                              'images/themes/theme9/bg.gif',
        'profile_link_color': '2FC2EF',
        'profile_image_url': 'http://a0.twimg.com/profile_images/532018826/'
                             'n587119531_1939735_9305_normal.jpg',
        'is_translator': False,
        'geo_enabled': False,
        'id_str': '74313638',
        'profile_background_image_url': 'http://a0.twimg.com/images/themes/'
                                        'theme9/bg.gif',
        'screen_name': 'foobar',
        'lang': 'en',
        'profile_background_tile': False,
        'favourites_count': 2,
        'name': 'Foo',
        'notifications': False,
        'url': None,
        'created_at': 'Tue Sep 15 00:26:17 +0000 2009',
        'contributors_enabled': False,
        'time_zone': 'Buenos Aires',
        'protected': False,
        'default_profile': False,
        'following': False
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class TwitterOAuth1IncludeEmailTest(OAuth1Test):
    backend_path = 'social_core.backends.twitter.TwitterOAuth'
    user_data_url = 'https://api.twitter.com/1.1/account/' \
        'verify_credentials.json?include_email=true'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    request_token_body = urlencode({
        'oauth_token_secret': 'foobar-secret',
        'oauth_token': 'foobar',
        'oauth_callback_confirmed': 'true'
    })
    user_data_body = json.dumps({
        'follow_request_sent': False,
        'profile_use_background_image': True,
        'id': 10101010,
        'description': 'Foo bar baz qux',
        'verified': False,
        'entities': {
            'description': {
                'urls': []
            }
        },
        'profile_image_url_https': 'https://twimg0-a.akamaihd.net/'
                                   'profile_images/532018826/'
                                   'n587119531_1939735_9305_normal.jpg',
        'profile_sidebar_fill_color': '252429',
        'profile_text_color': '666666',
        'followers_count': 77,
        'profile_sidebar_border_color': '181A1E',
        'location': 'Fooland',
        'default_profile_image': False,
        'listed_count': 4,
        'status': {
            'favorited': False,
            'contributors': None,
            'retweeted_status': {
                'favorited': False,
                'contributors': None,
                'truncated': False,
                'source': 'web',
                'text': '"Foo foo foo foo',
                'created_at': 'Fri Dec 21 18:12:00 +0000 2012',
                'retweeted': True,
                'in_reply_to_status_id': None,
                'coordinates': None,
                'id': 101010101010101010,
                'entities': {
                    'user_mentions': [],
                    'hashtags': [],
                    'urls': []
                },
                'in_reply_to_status_id_str': None,
                'place': None,
                'id_str': '101010101010101010',
                'in_reply_to_screen_name': None,
                'retweet_count': 8,
                'geo': None,
                'in_reply_to_user_id_str': None,
                'in_reply_to_user_id': None
            },
            'truncated': False,
            'source': 'web',
            'text': 'RT @foo: "Foo foo foo foo',
            'created_at': 'Fri Dec 21 18:24:10 +0000 2012',
            'retweeted': True,
            'in_reply_to_status_id': None,
            'coordinates': None,
            'id': 101010101010101010,
            'entities': {
                'user_mentions': [{
                    'indices': [3, 10],
                    'id': 10101010,
                    'screen_name': 'foo',
                    'id_str': '10101010',
                    'name': 'Foo'
                }],
                'hashtags': [],
                'urls': []
            },
            'in_reply_to_status_id_str': None,
            'place': None,
            'id_str': '101010101010101010',
            'in_reply_to_screen_name': None,
            'retweet_count': 8,
            'geo': None,
            'in_reply_to_user_id_str': None,
            'in_reply_to_user_id': None
        },
        'utc_offset': -10800,
        'statuses_count': 191,
        'profile_background_color': '1A1B1F',
        'friends_count': 151,
        'profile_background_image_url_https': 'https://twimg0-a.akamaihd.net/'
                                              'images/themes/theme9/bg.gif',
        'profile_link_color': '2FC2EF',
        'profile_image_url': 'http://a0.twimg.com/profile_images/532018826/'
                             'n587119531_1939735_9305_normal.jpg',
        'is_translator': False,
        'geo_enabled': False,
        'id_str': '74313638',
        'profile_background_image_url': 'http://a0.twimg.com/images/themes/'
                                        'theme9/bg.gif',
        'screen_name': 'foobar',
        'lang': 'en',
        'profile_background_tile': False,
        'favourites_count': 2,
        'name': 'Foo',
        'notifications': False,
        'url': None,
        'created_at': 'Tue Sep 15 00:26:17 +0000 2009',
        'contributors_enabled': False,
        'time_zone': 'Buenos Aires',
        'protected': False,
        'default_profile': False,
        'following': False,
        'email': 'foo@bar.bas'
    })

    def test_login(self):
        user = self.do_login()
        self.assertEqual(user.email, 'foo@bar.bas')

    def test_partial_pipeline(self):
        self.do_partial_pipeline()




import json
import os
import re
import sys
import unittest
from os import path
from urllib.parse import urlparse, urlunparse, urlencode, parse_qs

import requests

from unittest.mock import patch
from httpretty import HTTPretty

try:
    from onelogin.saml2.utils import OneLogin_Saml2_Utils
except ImportError:
    # Only available for python 2.7 at the moment, so don't worry if this fails
    pass

from .base import BaseBackendTest
from ...exceptions import AuthMissingParameter


DATA_DIR = path.join(path.dirname(__file__), 'data')


@unittest.skipIf('TRAVIS' in os.environ,
                  'Travis-ci segfaults probably due to a bad '
                  'dependencies build')
@unittest.skipIf('__pypy__' in sys.builtin_module_names,
                  'dm.xmlsec not compatible with pypy')
class SAMLTest(BaseBackendTest):
    backend_path = 'social_core.backends.saml.SAMLAuth'
    expected_username = 'myself'

    def extra_settings(self):
        name = path.join(DATA_DIR, 'saml_config.json')
        with open(name, 'r') as config_file:
            config_str = config_file.read()
        return json.loads(config_str)

    def setUp(self):
        """Patch the time so that we can replay canned
        request/response pairs"""
        super().setUp()

        @staticmethod
        def fixed_time():
            return OneLogin_Saml2_Utils.parse_SAML_to_time(
                '2015-05-09T03:57:22Z'
            )
        now_patch = patch.object(OneLogin_Saml2_Utils, 'now', fixed_time)
        now_patch.start()
        self.addCleanup(now_patch.stop)

    def install_http_intercepts(self, start_url, return_url):
        # When we request start_url
        # (https://idp.testshib.org/idp/profile/SAML2/Redirect/SSO...)
        # we will eventually get a redirect back, with SAML assertion
        # data in the query string.  A pre-recorded correct response
        # is kept in this .txt file:
        name = path.join(DATA_DIR, 'saml_response.txt')
        with open(name, 'r') as response_file:
            response_url = response_file.read()
        HTTPretty.register_uri(HTTPretty.GET, start_url, status=301,
                               location=response_url)
        HTTPretty.register_uri(HTTPretty.GET, return_url, status=200,
                               body='foobar')

    def do_start(self):
        start_url = self.backend.start().url
        # Modify the start URL to make the SAML request consistent
        # from test to test:
        start_url = self.modify_start_url(start_url)
        # If the SAML Identity Provider recognizes the user, we will
        # be redirected back to:
        return_url = self.backend.redirect_uri
        self.install_http_intercepts(start_url, return_url)
        response = requests.get(start_url)
        self.assertTrue(response.url.startswith(return_url))
        self.assertEqual(response.text, 'foobar')
        query_values = dict((k, v[0]) for k, v in
                            parse_qs(urlparse(response.url).query).items())
        self.assertNotIn(' ', query_values['SAMLResponse'])
        self.strategy.set_request_data(query_values, self.backend)
        return self.backend.complete()

    def test_metadata_generation(self):
        """Test that we can generate the metadata without error"""
        xml, errors = self.backend.generate_metadata_xml()
        self.assertEqual(len(errors), 0)
        self.assertEqual(xml.decode()[0], '<')

    def test_login(self):
        """Test that we can authenticate with a SAML IdP (TestShib)"""
        # pretend we've started with a URL like /login/saml/?idp=testshib:
        self.strategy.set_request_data({'idp': 'testshib'}, self.backend)
        self.do_login()

    def test_login_no_idp(self):
        """Logging in without an idp param should raise AuthMissingParameter"""
        with self.assertRaises(AuthMissingParameter):
            self.do_start()

    def modify_start_url(self, start_url):
        """
        Given a SAML redirect URL, parse it and change the ID to
        a consistent value, so the request is always identical.
        """
        # Parse the SAML Request URL to get the XML being sent to TestShib
        url_parts = urlparse(start_url)
        query = dict((k, v[0]) for (k, v) in
                     parse_qs(url_parts.query).items())
        xml = OneLogin_Saml2_Utils.decode_base64_and_inflate(
            query['SAMLRequest']
        )
        # Modify the XML:
        xml = xml.decode()
        xml, changed = re.subn(r'ID="[^"]+"', 'ID="TEST_ID"', xml)
        self.assertEqual(changed, 1)
        # Update the URL to use the modified query string:
        query['SAMLRequest'] = OneLogin_Saml2_Utils.deflate_and_base64_encode(
            xml
        )
        url_parts = list(url_parts)
        url_parts[4] = urlencode(query)
        return urlunparse(url_parts)



import json

from .oauth import OAuth2Test


class QuizletOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.quizlet.QuizletOAuth2'
    expected_username = 'foo_bar'

    access_token_body = json.dumps({
        'access_token': 'EE1IDxytP04tJ767GbjH7ED9PpGmYvL',
        'token_type': 'Bearer',
        'expires_in': 3600,
        'scope': 'read',
        'user_id': 'foo_bar'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Pocket OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/pocket.html
"""
from .base import BaseAuth
from ..utils import handle_http_errors


class PocketAuth(BaseAuth):
    name = 'pocket'
    AUTHORIZATION_URL = 'https://getpocket.com/auth/authorize'
    ACCESS_TOKEN_URL = 'https://getpocket.com/v3/oauth/authorize'
    REQUEST_TOKEN_URL = 'https://getpocket.com/v3/oauth/request'
    ID_KEY = 'username'

    def get_json(self, url, *args, **kwargs):
        headers = {'X-Accept': 'application/json'}
        kwargs.update({'method': 'POST', 'headers': headers})
        return super().get_json(url, *args, **kwargs)

    def get_user_details(self, response):
        return {'username': response['username']}

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        return response

    def auth_url(self):
        data = {
            'consumer_key': self.setting('KEY'),
            'redirect_uri': self.redirect_uri,
        }
        token = self.get_json(self.REQUEST_TOKEN_URL, data=data)['code']
        self.strategy.session_set('pocket_request_token', token)
        bits = (self.AUTHORIZATION_URL, token, self.redirect_uri)
        return '%s?request_token=%s&redirect_uri=%s' % bits

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        data = {
            'consumer_key': self.setting('KEY'),
            'code': self.strategy.session_get('pocket_request_token'),
        }
        response = self.get_json(self.ACCESS_TOKEN_URL, data=data)
        kwargs.update({'response': response, 'backend': self})
        return self.strategy.authenticate(*args, **kwargs)



import json
from urllib.parse import urlencode

import requests
from httpretty import HTTPretty

from .oauth import OAuth1Test


class YahooOAuth1Test(OAuth1Test):
    backend_path = 'social_core.backends.yahoo.YahooOAuth'
    user_data_url = 'https://social.yahooapis.com/v1/user/a-guid/profile?' \
                    'format=json'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    request_token_body = urlencode({
        'oauth_token_secret': 'foobar-secret',
        'oauth_token': 'foobar',
        'oauth_callback_confirmed': 'true'
    })
    guid_body = json.dumps({
        'guid': {
            'uri': 'https://social.yahooapis.com/v1/me/guid',
            'value': 'a-guid'
        }
    })
    user_data_body = json.dumps({
        'profile': {
            'bdRestricted': True,
            'memberSince': '2007-12-11T14:40:30Z',
            'image': {
                'width': 192,
                'imageUrl': 'http://l.yimg.com/dh/ap/social/profile/'
                            'profile_b192.png',
                'size': '192x192',
                'height': 192
            },
            'created': '2013-03-18T04:15:08Z',
            'uri': 'https://social.yahooapis.com/v1/user/a-guid/profile',
            'isConnected': False,
            'profileUrl': 'http://profile.yahoo.com/a-guid',
            'guid': 'a-guid',
            'nickname': 'foobar',
            'emails': [{
                'handle': 'foobar@yahoo.com',
                'id': 1,
                'primary': True,
                'type': 'HOME',
            }, {
                'handle': 'foobar@email.com',
                'id': 2,
                'type': 'HOME',
            }],
        }
    })

    def test_login(self):
        HTTPretty.register_uri(
            HTTPretty.GET,
            'https://social.yahooapis.com/v1/me/guid?format=json',
            status=200,
            body=self.guid_body
        )
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def test_get_user_details(self):
        HTTPretty.register_uri(
            HTTPretty.GET,
            self.user_data_url,
            status=200,
            body=self.user_data_body
        )
        response = requests.get(self.user_data_url)
        user_details = self.backend.get_user_details(
            response.json()['profile']
        )
        self.assertEqual(user_details['email'], 'foobar@yahoo.com')



import json

from .oauth import OAuth2Test


class SimpleLoginOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.simplelogin.SimpleLoginOAuth2'
    user_data_url = 'https://app.simplelogin.io/oauth2/userinfo'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'client': 'Continental',
        'email': 'john@wick.com',
        'email_verified': True,
        'id': 1,
        'name': 'John Wick',
        'avatar_url': 'http://wick.com/john.png'
    })
    expected_username = 'john@wick.com'

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from urllib.parse import urlencode

from .oauth import BaseOAuth2


class SalesforceOAuth2(BaseOAuth2):
    """Salesforce OAuth2 authentication backend"""
    name = 'salesforce-oauth2'
    AUTHORIZATION_URL = \
        'https://login.salesforce.com/services/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://login.salesforce.com/services/oauth2/token'
    REVOKE_TOKEN_URL = 'https://login.salesforce.com/services/oauth2/revoke'
    ACCESS_TOKEN_METHOD = 'POST'
    REFRESH_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ' '
    EXTRA_DATA = [
        ('id', 'id'),
        ('instance_url', 'instance_url'),
        ('issued_at', 'issued_at'),
        ('signature', 'signature'),
        ('refresh_token', 'refresh_token'),
    ]

    def get_user_details(self, response):
        """Return user details from a Salesforce account"""
        return {
            'username': response.get('username'),
            'email': response.get('email') or '',
            'first_name': response.get('first_name'),
            'last_name': response.get('last_name'),
            'fullname': response.get('display_name')
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        user_id_url = kwargs.get('response').get('id')
        url = user_id_url + '?' + urlencode({'access_token': access_token})
        try:
            return self.get_json(url)
        except ValueError:
            return None


class SalesforceOAuth2Sandbox(SalesforceOAuth2):
    """Salesforce OAuth2 authentication testing backend"""
    name = 'salesforce-oauth2-sandbox'
    AUTHORIZATION_URL = 'https://test.salesforce.com/services/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://test.salesforce.com/services/oauth2/token'
    REVOKE_TOKEN_URL = 'https://test.salesforce.com/services/oauth2/revoke'



import json
from urllib.parse import urlencode

from .oauth import OAuth1Test


class ThisIsMyJameOAuth1Test(OAuth1Test):
    backend_path = 'social_core.backends.thisismyjam.ThisIsMyJamOAuth1'
    user_data_url = 'http://api.thisismyjam.com/1/verify.json'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    request_token_body = urlencode({
        'oauth_token_secret': 'foobar-secret',
        'oauth_token': 'foobar',
        'oauth_callback_confirmed': 'true'
    })
    user_data_body = json.dumps({
        'id': 10101010,
        'person': {
            'name': 'foobar',
            'fullname': 'Foo Bar'
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from social_core.backends.oauth import BaseOAuth2


class CognitoOAuth2(BaseOAuth2):
    name = 'cognito'
    ID_KEY = 'username'
    DEFAULT_SCOPE = ['openid', 'profile', 'email']
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False

    def user_pool_domain(self):
        return self.setting('POOL_DOMAIN')

    def authorization_url(self):
        return '{}/login'.format(self.user_pool_domain())

    def access_token_url(self):
        return '{}/oauth2/token'.format(self.user_pool_domain())

    def user_data_url(self):
        return '{}/oauth2/userInfo'.format(self.user_pool_domain())

    def get_user_details(self, response):
        """Return user details from their cognito pool account"""
        first_name = response.get('given_name') or ''
        last_name = response.get('family_name') or ''
        fullname, first_name, last_name = self.get_user_names(
            first_name=first_name,
            last_name=last_name,
        )
        return {'username': response.get('username') or response.get('email'),
                'email': response.get('email'),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Grab user profile information from cognito."""
        response = self.get_json(
            url=self.user_data_url(),
            headers={'Authorization': 'Bearer {}'.format(access_token)},
        )

        user_data = {
            'given_name': response.get('given_name'),
            'family_name': response.get('family_name'),
            'username': response.get('username'),
            'email': response.get('email'),
        }

        return user_data



import json

from .oauth import OAuth2Test

from social_core.backends.orcid import ORCIDOAuth2

class ORCIDOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.orcid.ORCIDOAuth2'
    user_data_url = ORCIDOAuth2.USER_ID_URL
    expected_username = '0000-0002-2601-8132'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer',
        'orcid-identifier': {
            'path': '0000-0002-2601-8132'
        },
    })
    user_data_body = json.dumps({
        'orcid-identifier': {
            'uri': 'http://orcid.org/0000-0002-2601-8132',
            'path': '0000-0002-2601-8132',
            'host': 'orcid.org'
        },
        'person': {
            'last-modified-date': None,
            'name': {
                'created-date': {
                    'value': 1578249746904
                },
                'last-modified-date': {
                    'value': 1578249746904
                },
                'given-names': {
                    'value': 'Janani Kantharooban'
                },
                'family-name': {
                    'value': 'Umachanger'
                },
                'credit-name': None,
                'source': None,
                'visibility': 'PUBLIC',
                'path': '0000-0002-2601-8132'
            },
            'other-names': {
                'last-modified-date': None,
                'other-name': [],
                'path': '/0000-0002-2601-8132/other-names'
            },
            'biography': None,
            'researcher-urls': {
                'last-modified-date': None,
                'researcher-url': [],
                'path': '/0000-0002-2601-8132/researcher-urls'
            },
            'emails': {
                'last-modified-date': None,
                'email': [],
                'path': '/0000-0002-2601-8132/email'
            },
            'addresses': {
                'last-modified-date': None,
                'address': [],
                'path': '/0000-0002-2601-8132/address'
            },
            'keywords': {
                'last-modified-date': None,
                'keyword': [],
                'path': '/0000-0002-2601-8132/keywords'
            },
            'external-identifiers': {
                'last-modified-date': None,
                'external-identifier': [],
                'path': '/0000-0002-2601-8132/external-identifiers'
            },
            'path': '/0000-0002-2601-8132/person'
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test


class EdmodoOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.edmodo.EdmodoOAuth2'
    user_data_url = 'https://api.edmodo.com/users/me'
    expected_username = 'foobar12345'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'username': 'foobar12345',
        'coppa_verified': False,
        'first_name': 'Foo',
        'last_name': 'Bar',
        'premium': False,
        'verified_institution_member': False,
        'url': 'https://api.edmodo.com/users/12345',
        'type': 'teacher',
        'time_zone': None,
        'end_level': None,
        'start_level': None,
        'locale': 'en',
        'subjects': None,
        'utc_offset': None,
        'email': 'foo.bar@example.com',
        'gender': None,
        'about': None,
        'user_title': None,
        'id': 12345,
        'avatars': {
            'small': 'https://api.edmodo.com/users/12345/avatar?type=small&u=5a15xug93m53mi4ey3ck4fvkq',
            'large': 'https://api.edmodo.com/users/12345/avatar?type=large&u=5a15xug93m53mi4ey3ck4fvkq'
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test


class ZoomOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.zoom.ZoomOAuth2'
    user_data_url = 'https://api.zoom.us/v2/users/me'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar-token',
        'token_type': 'bearer',
        'refresh_token': 'foobar-refresh-token',
        'expires_in': 3599,
        'scope': 'identity'
    })
    user_data_body = json.dumps({
        'id': 'foobar',
        'first_name': 'Foo',
        'last_name': 'Bar',
        'email': 'foobar@email.com',
        'type': 2,
        'role_name': 'Foobar',
        'pmi': 1234567890,
        'use_pmi': False,
        'vanity_url': 'https://foobar.zoom.us/my/foobar',
        'personal_meeting_url': 'https://foobar.zoom.us/j/1234567890',
        'timezone': 'America/Denver',
        'verified': 1,
        'dept': '',
        'created_at': '2019-04-05T15:24:32Z',
        'last_login_time': '2019-12-16T18:02:48Z',
        'last_client_version': 'version',
        'pic_url': 'https://foobar.zoom.us/p/123456789',
        'host_key': '123456',
        'jid': 'foobar@xmpp.zoom.us',
        'group_ids': [],
        'im_group_ids': [
            'foobar-group-id'
        ],
        'account_id': 'foobar-account-id',
        'language': 'en-US',
        'phone_country': 'US',
        'phone_number': '+1 1234567891',
        'status': 'active'
    })
    refresh_token_body = json.dumps({
        'access_token': 'foobar-new-token',
        'token_type': 'bearer',
        'refresh_token': 'foobar-new-refresh-token',
        'expires_in': 3599,
        'scope': 'identity'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def test_refresh_token(self):
        user, social = self.do_refresh_token()
        self.assertEqual(user.username, self.expected_username)
        self.assertEqual(social.extra_data['access_token'], 'foobar-new-token')



"""
Backend for SAML 2.0 support

Terminology:

"Service Provider" (SP): Your web app
"Identity Provider" (IdP): The third-party site that is authenticating
                           users via SAML
"""
from onelogin.saml2.auth import OneLogin_Saml2_Auth
from onelogin.saml2.settings import OneLogin_Saml2_Settings

from .base import BaseAuth
from ..exceptions import AuthFailed, AuthMissingParameter

# Helpful constants:
OID_COMMON_NAME = 'urn:oid:2.5.4.3'
OID_EDU_PERSON_PRINCIPAL_NAME = 'urn:oid:1.3.6.1.4.1.5923.1.1.1.6'
OID_EDU_PERSON_ENTITLEMENT = 'urn:oid:1.3.6.1.4.1.5923.1.1.1.7'
OID_GIVEN_NAME = 'urn:oid:2.5.4.42'
OID_MAIL = 'urn:oid:0.9.2342.19200300.100.1.3'
OID_SURNAME = 'urn:oid:2.5.4.4'
OID_USERID = 'urn:oid:0.9.2342.19200300.100.1.1'


class SAMLIdentityProvider:
    """Wrapper around configuration for a SAML Identity provider"""
    def __init__(self, name, **kwargs):
        """Load and parse configuration"""
        self.name = name
        # name should be a slug and must not contain a colon, which
        # could conflict with uid prefixing:
        assert ':' not in self.name and ' ' not in self.name, \
            'IdP "name" should be a slug (short, no spaces)'
        self.conf = kwargs

    def get_user_permanent_id(self, attributes):
        """
        The most important method: Get a permanent, unique identifier
        for this user from the attributes supplied by the IdP.

        If you want to use the NameID, it's available via
        attributes['name_id']
        """
        uid = attributes[self.conf.get('attr_user_permanent_id', OID_USERID)]
        if isinstance(uid, list):
            uid = uid[0]
        return uid

    # Attributes processing:
    def get_user_details(self, attributes):
        """
        Given the SAML attributes extracted from the SSO response, get
        the user data like name.
        """
        return {
            'fullname': self.get_attr(attributes, 'attr_full_name',
                                      OID_COMMON_NAME),
            'first_name': self.get_attr(attributes, 'attr_first_name',
                                        OID_GIVEN_NAME),
            'last_name': self.get_attr(attributes, 'attr_last_name',
                                       OID_SURNAME),
            'username': self.get_attr(attributes, 'attr_username',
                                      OID_USERID),
            'email': self.get_attr(attributes, 'attr_email',
                                   OID_MAIL),
        }

    def get_attr(self, attributes, conf_key, default_attribute):
        """
        Internal helper method.
        Get the attribute 'default_attribute' out of the attributes,
        unless self.conf[conf_key] overrides the default by specifying
        another attribute to use.
        """
        key = self.conf.get(conf_key, default_attribute)
        value = attributes[key] if key in attributes else None
        if isinstance(value, list):
            if len(value):
                value = value[0]
            else:
                value = None
        return value

    @property
    def entity_id(self):
        """Get the entity ID for this IdP"""
        # Required. e.g. "https://idp.testshib.org/idp/shibboleth"
        return self.conf['entity_id']

    @property
    def sso_url(self):
        """Get the SSO URL for this IdP"""
        # Required. e.g.
        # "https://idp.testshib.org/idp/profile/SAML2/Redirect/SSO"
        return self.conf['url']

    @property
    def slo_url(self):
        """Get the SLO URL for this IdP"""
        return self.conf.get('slo_url')

    @property
    def saml_config_dict(self):
        """Get the IdP configuration dict in the format required by
        python-saml"""
        result = {
            'entityId': self.entity_id,
            'singleSignOnService': {
                'url': self.sso_url,
                # python-saml only supports Redirect
                'binding': 'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect'
            },
        }

        if self.slo_url:
            result['singleLogoutService'] = {
                'url': self.slo_url,
                'binding': 'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect'
            }

        cert = self.conf.get('x509cert', None)
        if cert:
            result['x509cert'] = cert
            return result
        cert = self.conf.get('x509certMulti', None)
        if cert:
            result['x509certMulti'] = cert
            return result
        raise KeyError('IDP must contain x509cert or x509certMulti')


class DummySAMLIdentityProvider(SAMLIdentityProvider):
    """
    A placeholder IdP used when we must specify something, e.g. when
    generating SP metadata.

    If OneLogin_Saml2_Auth is modified to not always require IdP
    config, this can be removed.
    """
    def __init__(self):
        super().__init__(
            'dummy',
            entity_id='https://dummy.none/saml2',
            url='https://dummy.none/SSO',
            x509cert=''
        )


class SAMLAuth(BaseAuth):
    """
    PSA Backend that implements SAML 2.0 Service Provider (SP) functionality.

    Unlike all of the other backends, this one can be configured to work with
    many identity providers (IdPs). For example, a University that belongs to a
    Shibboleth federation may support authentication via ~100 partner
    universities. Also, the IdP configuration can be changed at runtime if you
    require that functionality - just subclass this and override `get_idp()`.

    Several settings are required. Here's an example:

    SOCIAL_AUTH_SAML_SP_ENTITY_ID = "https://saml.example.com/"
    SOCIAL_AUTH_SAML_SP_PUBLIC_CERT = "... X.509 certificate string ..."
    SOCIAL_AUTH_SAML_SP_PRIVATE_KEY = "... private key ..."
    SOCIAL_AUTH_SAML_ORG_INFO = {
        "en-US": {
            "name": "example",
            "displayname": "Example Inc.",
            "url": "http://example.com"
        }
    }
    SOCIAL_AUTH_SAML_TECHNICAL_CONTACT = {
        "givenName": "Tech Gal",
        "emailAddress": "technical@example.com"
    }
    SOCIAL_AUTH_SAML_SUPPORT_CONTACT = {
        "givenName": "Support Guy",
        "emailAddress": "support@example.com"
    }
    SOCIAL_AUTH_SAML_ENABLED_IDPS = {
        "testshib": {
            "entity_id": "https://idp.testshib.org/idp/shibboleth",
            "url": "https://idp.testshib.org/idp/profile/SAML2/Redirect/SSO",
            "x509cert": "MIIEDjCCAvagAwIBAgIBADANBgkqhkiG9w0B...
                        ...8Bbnl+ev0peYzxFyF5sQA==",
        }
    }

    Optional settings:
    SOCIAL_AUTH_SAML_SP_EXTRA = {}
    SOCIAL_AUTH_SAML_SECURITY_CONFIG = {}
    """
    name = 'saml'
    EXTRA_DATA = []

    def get_idp(self, idp_name):
        """Given the name of an IdP, get a SAMLIdentityProvider instance"""
        idp_config = self.setting('ENABLED_IDPS')[idp_name]
        return SAMLIdentityProvider(idp_name, **idp_config)

    def generate_saml_config(self, idp=None):
        """
        Generate the configuration required to instantiate OneLogin_Saml2_Auth
        """
        # The shared absolute URL that all IdPs redirect back to -
        # this is specified in our metadata.xml:
        abs_completion_url = self.redirect_uri
        config = {
            'contactPerson': {
                'technical': self.setting('TECHNICAL_CONTACT'),
                'support': self.setting('SUPPORT_CONTACT')
            },
            'debug': True,
            'idp': idp.saml_config_dict if idp else {},
            'organization': self.setting('ORG_INFO'),
            'security': {
                'metadataValidUntil': '',
                'metadataCacheDuration': 'P10D',  # metadata valid for ten days
            },
            'sp': {
                'assertionConsumerService': {
                    'url': abs_completion_url,
                    # python-saml only supports HTTP-POST
                    'binding': 'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST'
                },
                'entityId': self.setting('SP_ENTITY_ID'),
                'x509cert': self.setting('SP_PUBLIC_CERT'),
                'privateKey': self.setting('SP_PRIVATE_KEY'),
            },
            'strict': True,  # We must force strict mode - for security
        }
        config['security'].update(self.setting('SECURITY_CONFIG', {}))
        config['sp'].update(self.setting('SP_EXTRA', {}))
        return config

    def generate_metadata_xml(self):
        """
        Helper method that can be used from your web app to generate the XML
        metadata required to link your web app as a Service Provider.

        Returns (metadata XML string, list of errors)

        Example usage (Django):
            from ..apps.django_app.utils import load_strategy, \
                                                     load_backend
            def saml_metadata_view(request):
                complete_url = reverse('social:complete', args=("saml", ))
                saml_backend = load_backend(load_strategy(request), "saml",
                                            complete_url)
                metadata, errors = saml_backend.generate_metadata_xml()
                if not errors:
                    return HttpResponse(content=metadata,
                                        content_type='text/xml')
                return HttpResponseServerError(content=', '.join(errors))
        """
        config = self.generate_saml_config()
        saml_settings = OneLogin_Saml2_Settings(
            config,
            sp_validation_only=True
        )
        metadata = saml_settings.get_sp_metadata()
        errors = saml_settings.validate_metadata(metadata)
        return metadata, errors

    def _create_saml_auth(self, idp):
        """Get an instance of OneLogin_Saml2_Auth"""
        config = self.generate_saml_config(idp)
        request_info = {
            'https': 'on' if self.strategy.request_is_secure() else 'off',
            'http_host': self.strategy.request_host(),
            'script_name': self.strategy.request_path(),
            'server_port': self.strategy.request_port(),
            'get_data': self.strategy.request_get(),
            'post_data': self.strategy.request_post(),
        }
        return OneLogin_Saml2_Auth(request_info, config)

    def auth_url(self):
        """Get the URL to which we must redirect in order to
        authenticate the user"""
        try:
            idp_name = self.strategy.request_data()['idp']
        except KeyError:
            raise AuthMissingParameter(self, 'idp')
        auth = self._create_saml_auth(idp=self.get_idp(idp_name))
        # Below, return_to sets the RelayState, which can contain
        # arbitrary data.  We use it to store the specific SAML IdP
        # name, since we multiple IdPs share the same auth_complete
        # URL.
        return auth.login(return_to=idp_name)

    def get_user_details(self, response):
        """Get user details like full name, email, etc. from the
        response - see auth_complete"""
        idp = self.get_idp(response['idp_name'])
        return idp.get_user_details(response['attributes'])

    def get_user_id(self, details, response):
        """
        Get the permanent ID for this user from the response.
        We prefix each ID with the name of the IdP so that we can
        connect multiple IdPs to this user.
        """
        idp = self.get_idp(response['idp_name'])
        uid = idp.get_user_permanent_id(response['attributes'])
        return '{0}:{1}'.format(idp.name, uid)

    def auth_complete(self, *args, **kwargs):
        """
        The user has been redirected back from the IdP and we should
        now log them in, if everything checks out.
        """
        idp_name = self.strategy.request_data()['RelayState']
        idp = self.get_idp(idp_name)
        auth = self._create_saml_auth(idp)
        auth.process_response()
        errors = auth.get_errors()
        if errors or not auth.is_authenticated():
            reason = auth.get_last_error_reason()
            raise AuthFailed(
                self, 'SAML login failed: {0} ({1})'.format(errors, reason)
            )

        attributes = auth.get_attributes()
        attributes['name_id'] = auth.get_nameid()
        self._check_entitlements(idp, attributes)
        response = {
            'idp_name': idp_name,
            'attributes': attributes,
            'session_index': auth.get_session_index(),
        }
        kwargs.update({'response': response, 'backend': self})
        return self.strategy.authenticate(*args, **kwargs)

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        extra_data = super().extra_data(
            user, uid, response['attributes'], details=details, *args, **kwargs
        )
        extra_data['session_index'] = response['session_index']
        extra_data['name_id'] = response['attributes']['name_id']
        return extra_data

    def request_logout(self, idp_name, social_auth, return_to=None):
        idp = self.get_idp(idp_name)
        auth = self._create_saml_auth(idp)
        name_id = social_auth.extra_data['name_id']
        session_index = social_auth.extra_data['session_index']
        return auth.logout(name_id=name_id, session_index=session_index, return_to=return_to)

    def process_logout(self, idp_name, delete_session_cb):
        idp = self.get_idp(idp_name)
        auth = self._create_saml_auth(idp)
        url = auth.process_slo(delete_session_cb=delete_session_cb)
        errors = auth.get_errors()
        return url, errors

    def _check_entitlements(self, idp, attributes):
        """
        Additional verification of a SAML response before
        authenticating the user.

        Subclasses can override this method if they need custom
        validation code, such as requiring the presence of an
        eduPersonEntitlement.

        raise social_core.exceptions.AuthForbidden if the user should not
        be authenticated, or do nothing to allow the login pipeline to
        continue.
        """
        pass



import json
from urllib.parse import urlencode

from httpretty import HTTPretty

from ...actions import do_disconnect

from ..models import User
from .oauth import OAuth1Test, OAuth2Test
from .open_id_connect import OpenIdConnectTestMixin


class GoogleOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.google.GoogleOAuth2'
    user_data_url = 'https://www.googleapis.com/oauth2/v3/userinfo'
    expected_username = 'foo'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'profile': 'https://plus.google.com/101010101010101010101',
        'family_name': 'Bar',
        'sub': '101010101010101010101',
        'picture': 'https://lh5.googleusercontent.com/-ui-GqpNh5Ms/'
                   'AAAAAAAAAAI/AAAAAAAAAZw/a7puhHMO_fg/photo.jpg',
        'locale': 'en',
        'email_verified': True,
        'given_name': 'Foo',
        'email': 'foo@bar.com',
        'name': 'Foo Bar',
    })

    def test_login(self):
        self.do_login()
        last_request = HTTPretty.last_request
        self.assertEqual(last_request.method, 'GET')
        self.assertTrue(self.user_data_url.endswith(last_request.path))
        self.assertEqual(
            last_request.headers['Authorization'],
            'Bearer foobar',
        )
        self.assertEqual(last_request.querystring, {})

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def test_with_unique_user_id(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GOOGLE_OAUTH2_USE_UNIQUE_USER_ID': True,
        })
        self.do_login()


class GoogleOAuth1Test(OAuth1Test):
    backend_path = 'social_core.backends.google.GoogleOAuth'
    user_data_url = 'https://www.googleapis.com/userinfo/email'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    request_token_body = urlencode({
        'oauth_token_secret': 'foobar-secret',
        'oauth_token': 'foobar',
        'oauth_callback_confirmed': 'true'
    })
    user_data_body = urlencode({
        'email': 'foobar@gmail.com',
        'isVerified': 'true',
        'id': '101010101010101010101'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def test_with_unique_user_id(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GOOGLE_OAUTH_USE_UNIQUE_USER_ID': True
        })
        self.do_login()

    def test_with_anonymous_key_and_secret(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GOOGLE_OAUTH_KEY': None,
            'SOCIAL_AUTH_GOOGLE_OAUTH_SECRET': None
        })
        self.do_login()


class GoogleRevokeTokenTest(GoogleOAuth2Test):
    def test_revoke_token(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_GOOGLE_OAUTH2_REVOKE_TOKENS_ON_DISCONNECT': True
        })
        self.do_login()
        user = User.get(self.expected_username)
        user.password = 'password'
        HTTPretty.register_uri(self._method(self.backend.REVOKE_TOKEN_METHOD),
                               self.backend.REVOKE_TOKEN_URL,
                               status=200)
        do_disconnect(self.backend, user)


class GoogleOpenIdConnectTest(OpenIdConnectTestMixin, OAuth2Test):
    backend_path = \
        'social_core.backends.google_openidconnect.GoogleOpenIdConnect'
    user_data_url = \
        'https://www.googleapis.com/plus/v1/people/me/openIdConnect'
    issuer = 'accounts.google.com'
    openid_config_body = json.dumps({
        'issuer': 'https://accounts.google.com',
        'authorization_endpoint': 'https://accounts.google.com/o/oauth2/v2/auth',
        'token_endpoint': 'https://www.googleapis.com/oauth2/v4/token',
        'userinfo_endpoint': 'https://www.googleapis.com/oauth2/v3/userinfo',
        'revocation_endpoint': 'https://accounts.google.com/o/oauth2/revoke',
        'jwks_uri': 'https://www.googleapis.com/oauth2/v3/certs',
        'response_types_supported': [
            'code',
            'token',
            'id_token',
            'code token',
            'code id_token',
            'token id_token',
            'code token id_token',
            'none',
        ],
        'subject_types_supported': [
            'public',
        ],
        'id_token_signing_alg_values_supported': [
            'RS256',
        ],
        'scopes_supported': [
            'openid',
            'email',
            'profile',
        ],
        'token_endpoint_auth_methods_supported': [
            'client_secret_post',
            'client_secret_basic',
        ],
        'claims_supported': [
            'aud',
            'email',
            'email_verified',
            'exp',
            'family_name',
            'given_name',
            'iat',
            'iss',
            'locale',
            'name',
            'picture',
            'sub',
        ],
    })



import json

from .oauth import OAuth2Test


class InstagramOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.instagram.InstagramOAuth2'
    user_data_url = 'https://graph.instagram.com/me'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer',
        'meta': {
            'code': 200
        },
        'user': {
            'username': 'foobar',
            'id': '101010101'
        }
    })
    user_data_body = json.dumps({
        'meta': {
            'code': 200
        },
        'username': 'foobar',
        'id': '101010101'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test


class SpotifyOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.spotify.SpotifyOAuth2'
    user_data_url = 'https://api.spotify.com/v1/me'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'display_name': None,
        'external_urls': {
            'spotify': 'https://open.spotify.com/user/foobar'
        },
        'followers': {
            'href': None,
            'total': 0
        },
        'href': 'https://api.spotify.com/v1/users/foobar',
        'id': 'foobar',
        'images': [],
        'type': 'user',
        'uri': 'spotify:user:foobar'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
OpenStreetMap OAuth support.

This adds support for OpenStreetMap OAuth service. An application must be
registered first on OpenStreetMap and the settings
SOCIAL_AUTH_OPENSTREETMAP_KEY and SOCIAL_AUTH_OPENSTREETMAP_SECRET
must be defined with the corresponding values.

More info: https://wiki.openstreetmap.org/wiki/OAuth
"""
from xml.dom import minidom

from .oauth import BaseOAuth1


class OpenStreetMapOAuth(BaseOAuth1):
    """OpenStreetMap OAuth authentication backend"""
    name = 'openstreetmap'
    AUTHORIZATION_URL = 'https://www.openstreetmap.org/oauth/authorize'
    REQUEST_TOKEN_URL = 'https://www.openstreetmap.org/oauth/request_token'
    ACCESS_TOKEN_URL = 'https://www.openstreetmap.org/oauth/access_token'
    EXTRA_DATA = [
        ('id', 'id'),
        ('avatar', 'avatar'),
        ('account_created', 'account_created')
    ]

    def get_user_details(self, response):
        """Return user details from OpenStreetMap account"""
        return {
            'username': response['username'],
            'email': '',
            'fullname': '',
            'first_name': '',
            'last_name': ''
        }

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        response = self.oauth_request(
            access_token, 'https://api.openstreetmap.org/api/0.6/user/details'
        )
        try:
            dom = minidom.parseString(response.content)
        except ValueError:
            return None
        user = dom.getElementsByTagName('user')[0]
        try:
            avatar = dom.getElementsByTagName('img')[0].getAttribute('href')
        except IndexError:
            avatar = None
        return {
            'id': user.getAttribute('id'),
            'username': user.getAttribute('display_name'),
            'account_created': user.getAttribute('account_created'),
            'avatar': avatar
        }



"""
Github OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/github.html
"""
from urllib.parse import urljoin

from requests import HTTPError

from .oauth import BaseOAuth2
from ..exceptions import AuthFailed


class GithubOAuth2(BaseOAuth2):
    """Github OAuth authentication backend"""
    name = 'github'
    API_URL = 'https://api.github.com/'
    AUTHORIZATION_URL = 'https://github.com/login/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://github.com/login/oauth/access_token'
    ACCESS_TOKEN_METHOD = 'POST'
    SCOPE_SEPARATOR = ','
    REDIRECT_STATE = False
    STATE_PARAMETER = True
    SEND_USER_AGENT = True
    EXTRA_DATA = [
        ('id', 'id'),
        ('expires', 'expires'),
        ('login', 'login')
    ]

    def api_url(self):
        return self.API_URL

    def get_user_details(self, response):
        """Return user details from Github account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('name')
        )
        return {'username': response.get('login'),
                'email': response.get('email') or '',
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        data = self._user_data(access_token)
        if not data.get('email'):
            try:
                emails = self._user_data(access_token, '/emails')
            except (HTTPError, ValueError, TypeError):
                emails = []

            if emails:
                email = emails[0]
                primary_emails = [
                    e for e in emails
                    if not isinstance(e, dict) or e.get('primary')
                ]
                if primary_emails:
                    email = primary_emails[0]
                if isinstance(email, dict):
                    email = email.get('email', '')
                data['email'] = email
        return data

    def _user_data(self, access_token, path=None):
        url = urljoin(self.api_url(), 'user{0}'.format(path or ''))
        return self.get_json(url, headers={'Authorization': 'token {0}'.format(access_token)})


class GithubMemberOAuth2(GithubOAuth2):
    no_member_string = ''

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        user_data = super().user_data(access_token, *args, **kwargs)
        headers = {'Authorization': 'token {0}'.format(access_token)}
        try:
            self.request(self.member_url(user_data), headers=headers)
        except HTTPError as err:
            # if the user is a member of the organization, response code
            # will be 204, see http://bit.ly/ZS6vFl
            if err.response.status_code != 204:
                raise AuthFailed(self,
                                 'User doesn\'t belong to the organization')
        return user_data

    def member_url(self, user_data):
        raise NotImplementedError('Implement in subclass')


class GithubOrganizationOAuth2(GithubMemberOAuth2):
    """Github OAuth2 authentication backend for organizations"""
    name = 'github-org'
    no_member_string = 'User doesn\'t belong to the organization'

    def member_url(self, user_data):
        return urljoin(
            self.api_url(),
            'orgs/{org}/members/{username}'.format(
                org=self.setting('NAME'),
                username=user_data.get('login')
            )
        )


class GithubTeamOAuth2(GithubMemberOAuth2):
    """Github OAuth2 authentication backend for teams"""
    name = 'github-team'
    no_member_string = 'User doesn\'t belong to the team'

    def member_url(self, user_data):
        return urljoin(
            self.api_url(),
            'teams/{team_id}/members/{username}'.format(
                team_id=self.setting('ID'),
                username=user_data.get('login')
            )
        )


class GithubAppAuth(GithubOAuth2):
    """GitHub App OAuth authentication backend"""
    name = 'github-app'

    def validate_state(self):
        """
        Scenario 1: user clicks an icon/button on your website and initiates
                    social login. This works exacltly like standard OAuth and we
                    have `state` and `redirect_uri`.

        Scenario 2: user starts from http://github.com/apps/your-app and clicks
                    'Install & Authorize' button! They still get a temporary
                    `code` (used to fetch `access_token`) but there's no `state`
                    or `redirect_uri` here.

        Note: Scenario 2 only happens when your GitHub App is configured
              with `Request user authorization (OAuth) during installation`
              turned on! This causes GitHub to redirect the person back to
              `/complete/github/`. If the above setting is turned off then
              GitHub will redirect to another URL called Setup URL and the
              person may need to login first before they can continue!
        """
        if self.data.get('installation_id') and self.data.get('setup_action'):
            return None

        return super().validate_state()



import json

from .oauth import OAuth2Test


class SoundcloudOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.soundcloud.SoundcloudOAuth2'
    user_data_url = 'https://api.soundcloud.com/me.json'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'website': None,
        'myspace_name': None,
        'public_favorites_count': 0,
        'followings_count': 0,
        'full_name': 'Foo Bar',
        'id': 10101010,
        'city': None,
        'track_count': 0,
        'playlist_count': 0,
        'discogs_name': None,
        'private_tracks_count': 0,
        'followers_count': 0,
        'online': True,
        'username': 'foobar',
        'description': None,
        'subscriptions': [],
        'kind': 'user',
        'quota': {
            'unlimited_upload_quota': False,
            'upload_seconds_left': 7200,
            'upload_seconds_used': 0
        },
        'website_title': None,
        'primary_email_confirmed': False,
        'permalink_url': 'http://soundcloud.com/foobar',
        'private_playlists_count': 0,
        'permalink': 'foobar',
        'upload_seconds_left': 7200,
        'country': None,
        'uri': 'https://api.soundcloud.com/users/10101010',
        'avatar_url': 'https://a1.sndcdn.com/images/'
                      'default_avatar_large.png?ca77017',
        'plan': 'Free'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test
from .open_id_connect import OpenIdConnectTestMixin


class FenceOpenIdConnectTest(OpenIdConnectTestMixin, OAuth2Test):
    backend_path = \
        'social_core.backends.fence.Fence'
    issuer = 'https://nci-crdc.datacommons.io/'
    openid_config_body = json.dumps({
        'issuer': 'https://nci-crdc.datacommons.io/',
        'authorization_endpoint': 'https://nci-crdc.datacommons.io/user/oauth2/authorize',
        'userinfo_endpoint': 'https://nci-crdc.datacommons.io/user/user/',
        'token_endpoint': 'https://nci-crdc.datacommons.io/user/oauth2/token',
        'revocation_endpoint': 'https://nci-crdc.datacommons.io/user/oauth2/revoke',
        'jwks_uri': 'https://auth.globus.org/jwk.json',
        'response_types_supported': [
            'code',
            'token',
            'token id_token',
            'id_token'
        ],
        'id_token_signing_alg_values_supported': [
            'RS512'
        ],
        'scopes_supported': [
            'ga4gh_passport_v1',
            'openid',
            'google_credentials',
            'google_service_account',
            'data',
            'user',
            'google_link',
            'admin',
            'fence'
        ],
        'token_endpoint_auth_methods_supported': [
            'authorization_code',
            'implicit'
        ],
        'claims_supported': [
            'aud',
            'sub',
            'iss',
            'exp',
            'jti',
            'auth_time',
            'azp',
            'nonce',
            'context'
        ],
        'subject_types_supported': ['public']
    })



from .oauth import BaseOAuth1


class WithingsOAuth(BaseOAuth1):
    name = 'withings'
    AUTHORIZATION_URL = 'https://developer.health.nokia.com/account/authorize'
    REQUEST_TOKEN_URL = 'https://developer.health.nokia.com/account/request_token'
    ACCESS_TOKEN_URL = 'https://developer.health.nokia.com/account/access_token'
    ID_KEY = 'userid'

    def get_user_details(self, response):
        """Return user details from Withings account"""
        return {'userid': response['access_token']['userid'],
                'email': ''}



from .oauth import BaseOAuth2


class ClasslinkOAuth(BaseOAuth2):
    """
    Classlink OAuth authentication backend.

    Docs: https://developer.classlink.com/docs/oauth2-workflow
    """
    name = 'classlink'
    AUTHORIZATION_URL = 'https://launchpad.classlink.com/oauth2/v2/auth'
    ACCESS_TOKEN_URL = 'https://launchpad.classlink.com/oauth2/v2/token'
    ACCESS_TOKEN_METHOD = 'POST'
    DEFAULT_SCOPE = ['profile']
    REDIRECT_STATE = False
    SCOPE_SEPARATOR = ' '

    def get_user_id(self, details, response):
        """Return user unique id provided by service"""
        return response['UserId']

    def get_user_details(self, response):
        """Return user details from Classlink account"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get('FirstName'),
            last_name=response.get('LastName')
        )

        return {
            'username': response.get('Email') or response.get('LoginId'),
            'email': response.get('Email'),
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name,
        }

    def user_data(self, token, *args, **kwargs):
        """Loads user data from service"""
        url = 'https://nodeapi.classlink.com/v2/my/info'
        auth_header = {'Authorization': 'Bearer %s' % token}
        try:
            return self.get_json(url, headers=auth_header)
        except ValueError:
            return None



import requests

from .oauth import BaseOAuth2
from ..exceptions import AuthFailed
from ..utils import handle_http_errors


class UntappdOAuth2(BaseOAuth2):
    """Untappd OAuth2 authentication backend"""
    name = 'untappd'
    AUTHORIZATION_URL = 'https://untappd.com/oauth/authenticate/'
    ACCESS_TOKEN_URL = 'https://untappd.com/oauth/authorize/'
    BASE_API_URL = 'https://api.untappd.com'
    USER_INFO_URL = BASE_API_URL + '/v4/user/info/'
    ACCESS_TOKEN_METHOD = 'GET'
    STATE_PARAMETER = False
    REDIRECT_STATE = False
    SEND_USER_AGENT = True
    EXTRA_DATA = [
        ('id', 'id'),
        ('bio', 'bio'),
        ('date_joined', 'date_joined'),
        ('location', 'location'),
        ('url', 'url'),
        ('user_avatar', 'user_avatar'),
        ('user_avatar_hd', 'user_avatar_hd'),
        ('user_cover_photo', 'user_cover_photo')
    ]

    def auth_params(self, state=None):
        client_id, client_secret = self.get_key_and_secret()
        params = {
            'client_id': client_id,
            'redirect_url': self.get_redirect_uri(),
            'response_type': self.RESPONSE_TYPE
        }
        return params

    def process_error(self, data):
        """
        All errors from Untappd are contained in the 'meta' key of the
        response.
        """
        response_code = data.get('meta', {}).get('http_code')
        if response_code is not None and response_code != requests.codes.ok:
            raise AuthFailed(self, data['meta']['error_detail'])

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        client_id, client_secret = self.get_key_and_secret()
        code = self.data.get('code')

        self.process_error(self.data)

        # Untapped sends the access token request with URL parameters,
        # not a body
        response = self.request_access_token(
            self.access_token_url(),
            method=self.ACCESS_TOKEN_METHOD,
            params={
                'response_type': 'code',
                'code': code,
                'client_id': client_id,
                'client_secret': client_secret,
                'redirect_url': self.get_redirect_uri()
            }
        )

        self.process_error(response)

        # Both the access_token and the rest of the response are
        # buried in the 'response' key
        return self.do_auth(
            response['response']['access_token'],
            response=response['response'],
            *args, **kwargs
        )

    def get_user_details(self, response):
        """Return user details from an Untappd account"""
        # Start with the user data as it was returned
        user_data = response['user']

        # Make a few updates to match expected key names
        user_data.update({
            'username': user_data.get('user_name'),
            'email': user_data.get('settings', {}).get('email_address', ''),
            'first_name': user_data.get('first_name'),
            'last_name': user_data.get('last_name'),
            'fullname': ' '.join([user_data.get('first_name'),
                                  user_data.get('last_name')])
        })
        return user_data

    def get_user_id(self, details, response):
        """
        Return a unique ID for the current user, by default from
        server response.
        """
        return response['user'].get(self.ID_KEY)

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        response = self.get_json(self.USER_INFO_URL, params={
            'access_token': access_token,
            'compact': 'true'
        })
        self.process_error(response)

        # The response data is buried in the 'response' key
        return response['response']



import json
from urllib.parse import urlencode

from .oauth import OAuth2Test


class StackoverflowOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.stackoverflow.StackoverflowOAuth2'
    user_data_url = 'https://api.stackexchange.com/2.1/me'
    expected_username = 'foobar'
    access_token_body = urlencode({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'items': [{
            'user_id': 101010,
            'user_type': 'registered',
            'creation_date': 1278525551,
            'display_name': 'foobar',
            'profile_image': 'http: //www.gravatar.com/avatar/'
                             '5280f15cedf540b544eecc30fcf3027c?'
                             'd=identicon&r=PG',
            'reputation': 547,
            'reputation_change_day': 0,
            'reputation_change_week': 0,
            'reputation_change_month': 0,
            'reputation_change_quarter': 65,
            'reputation_change_year': 65,
            'age': 22,
            'last_access_date': 1363544705,
            'last_modified_date': 1354035327,
            'is_employee': False,
            'link': 'http: //stackoverflow.com/users/101010/foobar',
            'location': 'Fooland',
            'account_id': 101010,
            'badge_counts': {
                'gold': 0,
                'silver': 3,
                'bronze': 6
            }
        }],
        'quota_remaining': 9997,
        'quota_max': 10000,
        'has_more': False
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import requests

from httpretty import HTTPretty

from ...actions import do_disconnect
from ...exceptions import NotAllowedToDisconnect
from ...utils import parse_qs

from ..models import User, TestUserSocialAuth
from .actions import BaseActionTest


class DisconnectActionTest(BaseActionTest):
    def test_not_allowed_to_disconnect(self):
        self.do_login()
        user = User.get(self.expected_username)
        with self.assertRaises(NotAllowedToDisconnect):
            do_disconnect(self.backend, user)

    def test_disconnect(self):
        self.do_login()
        user = User.get(self.expected_username)
        user.password = 'password'
        do_disconnect(self.backend, user)
        self.assertEqual(len(user.social), 0)

    def test_disconnect_with_association_id(self):
        self.do_login()
        user = User.get(self.expected_username)
        user.password = 'password'
        association_id = user.social[0].id
        second_usa = TestUserSocialAuth(user, user.social[0].provider, 'uid2')
        self.assertEqual(len(user.social), 2)
        do_disconnect(self.backend, user, association_id)
        self.assertEqual(len(user.social), 1)
        self.assertEqual(user.social[0], second_usa)

    def test_disconnect_with_partial_pipeline(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_DISCONNECT_PIPELINE': (
                'social_core.tests.pipeline.ask_for_password',
                'social_core.tests.pipeline.set_password',
                'social_core.pipeline.disconnect.allowed_to_disconnect',
                'social_core.pipeline.disconnect.get_entries',
                'social_core.pipeline.disconnect.revoke_tokens',
                'social_core.pipeline.disconnect.disconnect'
            )
        })
        self.do_login()
        user = User.get(self.expected_username)
        redirect = do_disconnect(self.backend, user)

        url = self.strategy.build_absolute_uri('/password')
        self.assertEqual(redirect.url, url)
        HTTPretty.register_uri(HTTPretty.GET, redirect.url, status=200,
                               body='foobar')
        HTTPretty.register_uri(HTTPretty.POST, redirect.url, status=200)

        password = 'foobar'
        requests.get(url)
        requests.post(url, data={'password': password})
        data = parse_qs(HTTPretty.last_request.body)
        self.assertEqual(data['password'], password)
        self.strategy.session_set('password', data['password'])

        redirect = do_disconnect(self.backend, user)
        self.assertEqual(len(user.social), 0)



import datetime
import json
from urllib.parse import urlencode

from httpretty import HTTPretty

from ...exceptions import AuthFailed

from .open_id import OpenIdTest


INFO_URL = 'http://api.steampowered.com/ISteamUser/GetPlayerSummaries/v0002/?'
JANRAIN_NONCE = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')


class SteamOpenIdTest(OpenIdTest):
    backend_path = 'social_core.backends.steam.SteamOpenId'
    expected_username = 'foobar'
    discovery_body = ''.join([
        '<?xml version="1.0" encoding="UTF-8"?>',
        '<xrds:XRDS xmlns:xrds="xri://$xrds" xmlns="xri://$xrd*($v*2.0)">',
        '<XRD>',
        '<Service priority="0">',
        '<Type>http://specs.openid.net/auth/2.0/server</Type>',
        '<URI>https://steamcommunity.com/openid/login</URI>',
        '</Service>',
        '</XRD>',
        '</xrds:XRDS>'
    ])
    user_discovery_body = ''.join([
        '<?xml version="1.0" encoding="UTF-8"?>',
        '<xrds:XRDS xmlns:xrds="xri://$xrds" xmlns="xri://$xrd*($v*2.0)">',
        '<XRD>',
        '<Service priority="0">',
        '<Type>http://specs.openid.net/auth/2.0/signon</Type>		',
        '<URI>https://steamcommunity.com/openid/login</URI>',
        '</Service>',
        '</XRD>',
        '</xrds:XRDS>'
    ])
    server_response = urlencode({
        'janrain_nonce': JANRAIN_NONCE,
        'openid.ns': 'http://specs.openid.net/auth/2.0',
        'openid.mode': 'id_res',
        'openid.op_endpoint': 'https://steamcommunity.com/openid/login',
        'openid.claimed_id': 'https://steamcommunity.com/openid/id/123',
        'openid.identity': 'https://steamcommunity.com/openid/id/123',
        'openid.return_to': 'http://myapp.com/complete/steam/?'
                            'janrain_nonce=' + JANRAIN_NONCE,
        'openid.response_nonce':
            JANRAIN_NONCE + 'oD4UZ3w9chOAiQXk0AqDipqFYRA=',
        'openid.assoc_handle': '1234567890',
        'openid.signed': 'signed,op_endpoint,claimed_id,identity,return_to,'
                         'response_nonce,assoc_handle',
        'openid.sig': '1az53vj9SVdiBwhk8%2BFQ68R2plo=',
    })
    player_details = json.dumps({
        'response': {
            'players': [{
                'steamid': '123',
                'primaryclanid': '1234',
                'timecreated': 1360768416,
                'personaname': 'foobar',
                'personastate': 0,
                'communityvisibilitystate': 3,
                'profileurl': 'http://steamcommunity.com/profiles/123/',
                'avatar': 'http://media.steampowered.com/steamcommunity/'
                          'public/images/avatars/fe/fef49e7fa7e1997310d7'
                          '05b2a6158ff8dc1cdfeb.jpg',
                'avatarfull': 'http://media.steampowered.com/steamcommunity/'
                              'public/images/avatars/fe/fef49e7fa7e1997310d7'
                              '05b2a6158ff8dc1cdfeb_full.jpg',
                'avatarmedium': 'http://media.steampowered.com/steamcommunity/'
                                'public/images/avatars/fe/fef49e7fa7e1997310d7'
                                '05b2a6158ff8dc1cdfeb_medium.jpg',
                'lastlogoff': 1360790014
            }]
        }
    })

    def _login_setup(self, user_url=None):
        self.strategy.set_settings({
            'SOCIAL_AUTH_STEAM_API_KEY': '123abc'
        })
        HTTPretty.register_uri(HTTPretty.POST,
                               'https://steamcommunity.com/openid/login',
                               status=200,
                               body=self.server_response)
        HTTPretty.register_uri(
            HTTPretty.GET,
            user_url or 'https://steamcommunity.com/openid/id/123',
            status=200,
            body=self.user_discovery_body
        )
        HTTPretty.register_uri(HTTPretty.GET,
                               INFO_URL,
                               status=200,
                               body=self.player_details)

    def test_login(self):
        self._login_setup()
        self.do_login()

    def test_partial_pipeline(self):
        self._login_setup()
        self.do_partial_pipeline()


class SteamOpenIdMissingSteamIdTest(SteamOpenIdTest):
    server_response = urlencode({
        'janrain_nonce': JANRAIN_NONCE,
        'openid.ns': 'http://specs.openid.net/auth/2.0',
        'openid.mode': 'id_res',
        'openid.op_endpoint': 'https://steamcommunity.com/openid/login',
        'openid.claimed_id': 'https://steamcommunity.com/openid/BROKEN',
        'openid.identity': 'https://steamcommunity.com/openid/BROKEN',
        'openid.return_to': 'http://myapp.com/complete/steam/?'
                            'janrain_nonce=' + JANRAIN_NONCE,
        'openid.response_nonce':
            JANRAIN_NONCE + 'oD4UZ3w9chOAiQXk0AqDipqFYRA=',
        'openid.assoc_handle': '1234567890',
        'openid.signed': 'signed,op_endpoint,claimed_id,identity,return_to,'
                         'response_nonce,assoc_handle',
        'openid.sig': '1az53vj9SVdiBwhk8%2BFQ68R2plo=',
    })

    def test_login(self):
        self._login_setup(user_url='https://steamcommunity.com/openid/BROKEN')
        with self.assertRaises(AuthFailed):
            self.do_login()

    def test_partial_pipeline(self):
        self._login_setup(user_url='https://steamcommunity.com/openid/BROKEN')
        with self.assertRaises(AuthFailed):
            self.do_partial_pipeline()
